# -*- coding: utf-8 -*-

from datetime import datetime

from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist
from django.db import models
from django.db.models import Max
from django.db.models.signals import post_save, post_delete, pre_delete, pre_save
from django.shortcuts import get_object_or_404
from django.urls import reverse
from uuslug import slugify

from tcms.core.models import TCMSActionModel
from tcms.core.utils import checksum
from tcms.core.tcms_router import connection
from tcms.management.models import Version
from tcms.testcases.models import TestCase
from tcms.testcases.models import TestCaseCategory
from tcms.testcases.models import TestCasePlan
from tcms.testcases.models import TestCaseStatus
from tcms.testplans import signals as plan_watchers

try:
    from tcms.core.contrib.plugins_support.signals import register_model
except ImportError:
    register_model = None


class TestPlanType(TCMSActionModel):
    id = models.AutoField(db_column='type_id', primary_key=True)
    name = models.CharField(max_length=64, unique=True)
    description = models.TextField(blank=True, null=True)

    def __str__(self):
        return self.name

    class Meta:
        db_table = 'test_plan_types'
        ordering = ['name']


class TestPlan(TCMSActionModel):
    """A plan within the TCMS"""

    plan_id = models.AutoField(max_length=11, primary_key=True)
    name = models.CharField(max_length=255, db_index=True)
    create_date = models.DateTimeField(db_column='creation_date', auto_now_add=True)
    is_active = models.BooleanField(db_column='isactive', default=True, db_index=True)
    extra_link = models.CharField(max_length=1024, default=None, blank=True, null=True)

    product_version = models.ForeignKey(Version, related_name='plans',
                                        on_delete=models.CASCADE)
    owner = models.ForeignKey('auth.User', blank=True, null=True,
                              related_name='myplans', on_delete=models.CASCADE)
    author = models.ForeignKey('auth.User', on_delete=models.CASCADE)
    product = models.ForeignKey('management.Product', related_name='plan',
                                on_delete=models.CASCADE)
    type = models.ForeignKey(TestPlanType, on_delete=models.CASCADE)
    parent = models.ForeignKey('self', blank=True, null=True,
                               related_name='child_set',
                               on_delete=models.SET_NULL)

    attachment = models.ManyToManyField('management.TestAttachment',
                                        through='testplans.TestPlanAttachment')
    component = models.ManyToManyField('management.Component',
                                       through='testplans.TestPlanComponent')
    env_group = models.ManyToManyField('management.TCMSEnvGroup',
                                       related_name='plans',
                                       through='TCMSEnvPlanMap')
    tag = models.ManyToManyField('management.TestTag', through='testplans.TestPlanTag')

    class Meta:
        db_table = 'test_plans'
        index_together = [['product', 'plan_id']]

    def __str__(self):
        return self.name

    @classmethod
    def to_xmlrpc(cls, query=None):
        from tcms.xmlrpc.serializer import TestPlanXMLRPCSerializer
        from tcms.xmlrpc.utils import distinct_filter

        _query = query or {}
        qs = distinct_filter(TestPlan, _query).order_by('pk')
        s = TestPlanXMLRPCSerializer(model_class=cls, queryset=qs)
        return s.serialize_queryset()

    @classmethod
    def list(cls, query=None):
        '''docstring for list_plans'''
        from django.db.models import Q

        new_query = {}

        for k, v in query.items():
            if v and k not in ['action', 't', 'f', 'a']:
                new_query[k] = hasattr(v, 'strip') and v.strip() or v

        # build a QuerySet:
        q = cls.objects
        # add any necessary filters to the query:

        if new_query.get('search'):
            q = q.filter(Q(plan_id__icontains=new_query['search']) |
                         Q(name__icontains=new_query['search']))
            del new_query['search']

        return q.filter(**new_query).distinct()

    def confirmed_case(self):
        return self.case.filter(case_status__name='CONFIRMED')

    def latest_text(self):
        try:
            return self.text.select_related('author').order_by(
                '-plan_text_version')[0]
        except IndexError:
            return None
        except ObjectDoesNotExist:
            return None

    def text_exist(self):
        try:
            return self.text.exists()
        except IndexError:
            return False
        except ObjectDoesNotExist:
            return False

    def text_checksum(self):
        try:
            return self.text.order_by('-plan_text_version').only(
                'checksum')[0].checksum
        except IndexError:
            return None
        except ObjectDoesNotExist:
            return None

    def get_text_with_version(self, plan_text_version=None):
        if plan_text_version:
            try:
                return self.text.get(
                    plan_text_version=plan_text_version
                )
            except TestPlanText.DoesNotExist:
                return None

        return self.latest_text()

    def add_text(self,
                 author,
                 plan_text,
                 create_date=datetime.now(),
                 plan_text_version=None,
                 text_checksum=None):
        if not plan_text_version:
            latest_text = self.latest_text()
            if latest_text:
                plan_text_version = latest_text.plan_text_version + 1
            else:
                plan_text_version = 1

        if not text_checksum:
            old_checksum = self.text_checksum()
            new_checksum = checksum(plan_text)
            if old_checksum == new_checksum:
                return self.latest_text()

        return self.text.create(
            plan_text_version=plan_text_version,
            author=author,
            create_date=create_date,
            plan_text=plan_text,
            checksum=text_checksum or checksum(plan_text)
        )

    def add_case(self, case, sortkey=1):
        """Add a case"""
        tcp, is_created = TestCasePlan.objects.get_or_create(
            plan=self,
            case=case,
        )
        if is_created:
            tcp.sortkey = sortkey
            tcp.save(update_fields=['sortkey'])

    def add_component(self, component):
        try:
            return TestPlanComponent.objects.create(
                plan=self,
                component=component,
            )
        except Exception:
            return False

    def add_env_group(self, env_group):
        # Create the env plan map
        return TCMSEnvPlanMap.objects.create(
            plan=self,
            group=env_group,
        )

    def add_attachment(self, attachment):
        return TestPlanAttachment.objects.create(
            plan=self,
            attachment=attachment,
        )

    def add_tag(self, tag):
        return TestPlanTag.objects.get_or_create(
            plan=self,
            tag=tag
        )

    def remove_tag(self, tag):
        cursor = connection.writer_cursor
        cursor.execute(
            "DELETE from test_plan_tags WHERE plan_id = %s AND tag_id = %s",
            (self.pk, tag.pk))

    def remove_component(self, component):
        try:
            return TestPlanComponent.objects.get(
                plan=self, component=component
            ).delete()
        except Exception:
            return False

    def clear_env_groups(self):
        # Remove old env groups because we only maintanence on group per plan.
        return TCMSEnvPlanMap.objects.filter(plan=self).delete()

    def delete_case(self, case):
        TestCasePlan.objects.filter(case=case.pk, plan=self.pk).delete()

    def get_absolute_url(self):
        return reverse('plan-get', kwargs={
            'plan_id': self.plan_id,
            'slug': slugify(self.name),
        })

    def get_case_sortkey(self):
        '''
        Get case sortkey.
        '''
        result = TestCasePlan.objects.filter(plan=self).aggregate(Max('sortkey'))
        sortkey = result['sortkey__max']
        if sortkey is None:
            return None
        else:
            return sortkey + 10

    def make_cloned_name(self):
        """Make default name of cloned plan"""
        return f'Copy of {self.name}'

    def clone(self, new_name=None, product=None, version=None,
              new_original_author=None, set_parent=True,
              copy_texts=True, default_text_author=None,
              copy_attachments=True,
              copy_environment_group=True,
              link_cases=True, copy_cases=None,
              new_case_author=None,
              new_case_default_tester=None,
              default_component_initial_owner=None):
        """Clone this plan

        :param str new_name: New name of cloned plan. If not passed, make_cloned_name is called
            to generate a default one.
        :param product: Product of cloned plan. If not passed, original plan's product is used.
        :param version: Product version of cloned plan. If not passed, original plan's
            product_version is used.
        :param new_original_author: New author of cloned plan. If not passed, original plan's
            author is used.
        :param bool set_parent: Whether to set original plan as parent of cloned plan.
            Set by default.
        :param bool copy_texts: Whether to copy the four text. Copy by default.
        :param default_text_author: When not copy the four text, new text will be created.
            This is the default author of new created text.
        :param bool copy_attachments: Whether to copy attachments. Copy by default.
        :param bool copy_environment_group: Whether to copy environment groups. Copy by default.
        :param bool link_cases: Whether to link cases to cloned plan. Default is True.
        :param bool copy_cases: Whether to copy cases to cloned plan instead of just linking them.
            Default is False.
        :param new_case_author: The author of copied cases. Used only if copy cases.
        :param new_case_default_tester: The default tester of copied cases. Used only if copy cases.
        :param default_component_initial_owner: Used only if copy cases. If copied case does not
            have original case' component, create it and use this value as the initial_owner.
        :rtype: cloned plan
        """

        if not copy_texts and not default_text_author:
            raise ValueError('Missing default text author when not copy texts.')

        if not copy_cases and not default_component_initial_owner:
            raise ValueError('Missing default component initial owner when not copy cases.')

        tp_dest = TestPlan.objects.create(
            name=new_name or self.make_cloned_name(),
            product=product or self.product,
            author=new_original_author or self.author,
            type=self.type,
            product_version=version or self.product_version,
            create_date=self.create_date,
            is_active=self.is_active,
            extra_link=self.extra_link,
            parent=self if set_parent else None)

        # Copy the plan documents
        if copy_texts:
            tptxts_src = self.text.all()
            for tptxt_src in tptxts_src:
                tp_dest.add_text(
                    plan_text_version=tptxt_src.plan_text_version,
                    author=tptxt_src.author,
                    create_date=tptxt_src.create_date,
                    plan_text=tptxt_src.plan_text)
        else:
            tp_dest.add_text(author=default_text_author, plan_text='')

        # Copy the plan tags
        for tp_tag_src in self.tag.all():
            tp_dest.add_tag(tag=tp_tag_src)

        # Copy the plan attachments
        if copy_attachments:
            for tp_attach_src in self.attachment.all():
                tp_dest.add_attachment(attachment=tp_attach_src)

        # Copy the environment group
        if copy_environment_group:
            for env_group in self.env_group.all():
                tp_dest.add_env_group(env_group=env_group)

        # Link the cases of the plan
        if link_cases:
            tpcases_src = self.case.all()

            if copy_cases:
                for tpcase_src in tpcases_src:
                    tcp = get_object_or_404(TestCasePlan, plan=self, case=tpcase_src)
                    author = new_case_author or tpcase_src.author
                    default_tester = new_case_default_tester or tpcase_src.default_tester

                    tc_category, b_created = TestCaseCategory.objects.get_or_create(
                        name=tpcase_src.category.name, product=product)

                    tpcase_dest = TestCase.objects.create(
                        create_date=tpcase_src.create_date,
                        is_automated=tpcase_src.is_automated,
                        script=tpcase_src.script,
                        arguments=tpcase_src.arguments,
                        summary=tpcase_src.summary,
                        requirement=tpcase_src.requirement,
                        alias=tpcase_src.alias,
                        estimated_time=tpcase_src.estimated_time,
                        case_status=TestCaseStatus.get('PROPOSED'),
                        category=tc_category,
                        priority=tpcase_src.priority,
                        author=author,
                        default_tester=default_tester)

                    # Add case to plan.
                    tp_dest.add_case(tpcase_dest, tcp.sortkey)

                    for tc_tag_src in tpcase_src.tag.all():
                        tpcase_dest.add_tag(tag=tc_tag_src)

                    for component in tpcase_src.component.filter(product__id=self.product_id):
                        try:
                            new_c = tp_dest.product.component.get(name=component.name)
                        except ObjectDoesNotExist:
                            new_c = tp_dest.product.component.create(
                                name=component.name,
                                initial_owner=default_component_initial_owner,
                                description=component.description)

                        tpcase_dest.add_component(new_c)

                    text = tpcase_src.latest_text()

                    if text:
                        tpcase_dest.add_text(author=text.author,
                                             action=text.action,
                                             effect=text.effect,
                                             setup=text.setup,
                                             breakdown=text.breakdown,
                                             create_date=text.create_date)
            else:
                for tpcase_src in tpcases_src:
                    tcp = get_object_or_404(TestCasePlan, plan=self, case=tpcase_src)
                    tp_dest.add_case(tpcase_src, tcp.sortkey)

        return tp_dest

    def import_cases(self, cases_info, sortkey_step=10):
        """Import a list of cases

        :param cases_info: list of mappings, each of which contains data to
            create a case. This information is created from case XML document
            which is exported previously from Nitrate.
        :type cases_info: list[dict]
        """
        sortkey = 1
        for info in cases_info:
            self._import_case(info, sortkey=sortkey)
            sortkey += sortkey_step

    def _import_case(self, case_info, sortkey):
        """Import a case

        :param dict case_info: refer to `import_cases`. This is one of the case
            information in the list.
        :param int sortkey: the sort key of imported case.
        """
        category, _ = TestCaseCategory.objects.get_or_create(
            product=self.product,
            name=case_info['category_name'])

        tc = TestCase.objects.create(
            is_automated=case_info['is_automated'],
            script='',
            arguments='',
            summary=case_info['summary'],
            requirement='',
            alias='',
            estimated_time=0,
            case_status_id=case_info['case_status_id'],
            category_id=category.id,
            priority_id=case_info['priority_id'],
            author_id=case_info['author_id'],
            default_tester_id=case_info['default_tester_id'],
            notes=case_info['notes'])

        tc.add_text(case_text_version=1,
                    author=case_info['author'],
                    action=case_info['action'],
                    effect=case_info['effect'],
                    setup=case_info['setup'],
                    breakdown=case_info['breakdown'])

        # handle tags
        if case_info['tags']:
            for tag in case_info['tags']:
                tc.add_tag(tag=tag)

        self.add_case(tc, sortkey=sortkey)


class TestPlanText(TCMSActionModel):
    plan = models.ForeignKey(TestPlan, related_name='text', on_delete=models.CASCADE)
    plan_text_version = models.IntegerField()
    author = models.ForeignKey('auth.User', db_column='who', on_delete=models.CASCADE)
    create_date = models.DateTimeField(auto_now_add=True,
                                       db_column='creation_ts')
    plan_text = models.TextField(blank=True)
    checksum = models.CharField(max_length=32)

    class Meta:
        db_table = 'test_plan_texts'
        ordering = ['plan', '-plan_text_version']
        unique_together = ('plan', 'plan_text_version')


class TestPlanAttachment(models.Model):
    attachment = models.ForeignKey('management.TestAttachment', on_delete=models.CASCADE)
    plan = models.ForeignKey(TestPlan, on_delete=models.CASCADE)

    class Meta:
        db_table = 'test_plan_attachments'


class TestPlanTag(models.Model):
    tag = models.ForeignKey('management.TestTag', on_delete=models.CASCADE)
    plan = models.ForeignKey(TestPlan, on_delete=models.CASCADE)
    user = models.IntegerField(default="1", db_column='userid')

    class Meta:
        db_table = 'test_plan_tags'


class TestPlanComponent(models.Model):
    plan = models.ForeignKey(TestPlan, on_delete=models.CASCADE)
    component = models.ForeignKey('management.Component', on_delete=models.CASCADE)

    class Meta:
        db_table = 'test_plan_components'
        unique_together = ('plan', 'component')


class TestPlanEmailSettings(models.Model):
    plan = models.OneToOneField(TestPlan, related_name='email_settings',
                                on_delete=models.CASCADE)
    is_active = models.BooleanField(default=False)
    auto_to_plan_owner = models.BooleanField(default=False)
    auto_to_plan_author = models.BooleanField(default=False)
    auto_to_case_owner = models.BooleanField(default=False)
    auto_to_case_default_tester = models.BooleanField(default=False)
    notify_on_plan_update = models.BooleanField(default=False)
    notify_on_plan_delete = models.BooleanField(default=False)
    notify_on_case_update = models.BooleanField(default=False)

    class Meta:
        pass


class TCMSEnvPlanMap(models.Model):
    plan = models.ForeignKey(TestPlan, on_delete=models.CASCADE)
    group = models.ForeignKey('management.TCMSEnvGroup', on_delete=models.CASCADE)

    class Meta:
        db_table = 'tcms_env_plan_map'


if register_model:
    register_model(TestPlan)
    register_model(TestPlanText)
    register_model(TestPlanType)
    register_model(TestPlanTag)
    register_model(TestPlanComponent)


def _listen():
    post_save.connect(plan_watchers.notify_on_plan_is_updated, TestPlan)
    pre_delete.connect(
        plan_watchers.load_email_settings_for_later_deletion, TestPlan)
    post_delete.connect(plan_watchers.notify_deletion_of_plan, TestPlan)
    pre_save.connect(plan_watchers.pre_save_clean, TestPlan)


def _disconnect_signals():
    """ used in testing """
    post_save.disconnect(plan_watchers.notify_on_plan_is_updated, TestPlan)
    pre_delete.disconnect(
        plan_watchers.load_email_settings_for_later_deletion, TestPlan)
    post_delete.disconnect(plan_watchers.notify_deletion_of_plan, TestPlan)
    pre_save.disconnect(plan_watchers.pre_save_clean, TestPlan)


if settings.LISTENING_MODEL_SIGNAL:
    _listen()



#!/usr/bin/python3

import argparse
import git
import re

parser = argparse.ArgumentParser()
parser.add_argument('--repo', help='Path to git repository.')
parser.add_argument('--since-version', required=False, help='Collect commits since this version.')
parser.add_argument('--link-issues', required=False, action='store_true',
                    help='Generate link markup for issues.')

args = parser.parse_args()

repo = git.Repo(args.repo or '.')

if args.since_version:
    last_version = args.since_version
else:
    last_version = repo.git.describe().split('-')[0]

issue_numbers = []

for commit in repo.iter_commits(f'HEAD...{last_version}'):
    match = re.search(r'Fixes #(\d+)', commit.message, re.MULTILINE)
    if match:
        if args.link_issues:
            issue_number = match.groups()[0]
            fixed_issue = f'- `#{issue_number}`_ '
            issue_numbers.append(issue_number)
        else:
            fixed_issue = '- #{} '.format(match.groups()[0])
    else:
        fixed_issue = ''
    print(f'* {commit.summary} {fixed_issue}({commit.author})')

if args.link_issues:
    print()
    for item in sorted(map(int, issue_numbers)):
        print(f'.. _#{item}: https://github.com/Nitrate/Nitrate/issues/{item}/')



# -*- coding: utf-8 -*-
from django import template

register = template.Library()


@register.filter(name='cutbystring')
def cut_by_string(value, arg):
    arg = int(arg)
    if len(value) < arg:
        return value
    else:
        return value[:arg - 3] + "..."


@register.filter(name='ismine')
def is_mine(object, user):
    if hasattr(object, 'author') and object.author == user:
        return True
    if hasattr(object, 'manager') and object.manager == user:
        return True

    return False


@register.filter(name='smart_unicode')
def smart_unicode(object):
    from django.utils.encoding import smart_unicode

    if not object:
        return object
    return smart_unicode(object)


@register.filter(name='smart_int')
def smart_int(object):
    if not object:
        return object

    try:
        return int(object)
    except ValueError:
        return object


@register.filter(name='timedelta2string')
def timedelta2string(object):
    from tcms.core.utils import format_timedelta
    return format_timedelta(object)


@register.filter(name='timedelta2seconds')
def timedelta2seconds(object):
    return int(object.seconds + object.days * 86400)


@register.inclusion_tag('mail/new_run.txt', takes_context=True)
def absolute_url(context):
    request = context['request']
    object = context['test_run']
    if not hasattr(object, 'get_absolute_url'):
        return None

    return object.get_absolute_url(request)


@register.filter()
def replace_str(value, arg):
    """Template filter to replace a substring"""
    old, new = arg.split(' ', 1)
    return value.replace(old, new)



# -*- coding: utf-8 -*-

from django.conf.urls import url
from . import views

urlpatterns = [
    url(r'^environment/groups/$', views.environment_groups,
        name='management-env-groups'),
    url(r'^environment/group/edit/$', views.environment_group_edit,
        name='management-env-group-edit'),
    url(r'^environment/properties/$', views.environment_properties,
        name='management-env-properties'),
    url(r'^environment/properties/values/$', views.environment_property_values,
        name='management-env-properties-values'),
]



# -*- coding: utf-8 -*-

from django.conf.urls import url

from tcms.testruns import views
from tcms.testruns import ajax

urlpatterns = [
    url(r'^new/$', views.new, name='run-new'),
    url(r'^(?P<run_id>\d+)/$', views.get, name='run-get'),
    url(r'^(?P<run_id>\d+)/clone/$', views.new_run_with_caseruns, name='run-clone'),
    url(r'^(?P<run_id>\d+)/delete/$', views.delete, name='run-delete'),
    url(r'^(?P<run_id>\d+)/edit/$', views.edit, name='run-edit'),

    url(r'^(?P<run_id>\d+)/report/$', views.TestRunReportView.as_view(),
        name='run-report'),

    url(r'^(?P<run_id>\d+)/ordercase/$', views.order_case, name='run-order-case'),
    url(r'^(?P<run_id>\d+)/changestatus/$', views.ChangeRunStatusView.as_view(),
        name='run-change-status'),
    url(r'^(?P<run_id>\d+)/ordercaserun/$', views.order_case, name='run-order-caserun'),
    url(r'^(?P<run_id>\d+)/removecaserun/$', views.RemoveCaseRunView.as_view(),
        name='run-remove-caserun'),

    url(r'^(?P<run_id>\d+)/assigncase/$', views.AddCasesToRunView.as_view(),
        name='add-cases-to-run'),

    url(r'^(?P<run_id>\d+)/cc/$', views.cc, name='run-cc'),
    url(r'^(?P<run_id>\d+)/update/$', views.update_case_run_text, name='run-update'),
    url(r'^(?P<run_id>\d+)/export/$', views.export, name='run-export'),

    url(r'^(?P<run_id>\d+)/case-run/(?P<case_run_id>\d+)/file-issue/$',
        views.FileIssueForCaseRun.as_view(), name='run-caserun-file-issue'),

    url(r'^(?P<run_id>\d+)/issues/$', ajax.manage_case_run_issues, name='run-issues'),
]



# -*- coding: utf-8 -*-

from django import http

__all__ = (
    'JsonResponseBadRequest',
    'JsonResponseServerError',
)


class JsonResponseBadRequest(http.JsonResponse):
    status_code = 400


class JsonResponseServerError(http.JsonResponse):
    status_code = 500



# -*- coding: utf-8 -*-

import os
import logging

from datetime import datetime

from django.conf import settings
from django.contrib.auth.mixins import PermissionRequiredMixin
from django.http import HttpResponse, HttpResponseRedirect, Http404
from django.http import JsonResponse
from django.shortcuts import get_object_or_404
from django.urls import reverse
from django.utils.encoding import smart_str
from django.views import generic
from django.views.decorators.http import require_GET
from six.moves.urllib_parse import unquote

from tcms.core.views import Prompt
from tcms.testcases.models import TestCase, TestCaseAttachment
from tcms.testplans.models import TestPlan, TestPlanAttachment
from tcms.management.models import TestAttachment, TestAttachmentData

log = logging.getLogger(__name__)


class UploadFileView(PermissionRequiredMixin, generic.View):
    """Upload a file"""

    permission_required = 'management.add_testattachment'

    def post(self, request):
        to_plan_id = request.POST.get('to_plan_id')
        to_case_id = request.POST.get('to_case_id')

        if to_plan_id is None and to_case_id is None:
            return Prompt.render(
                request=request,
                info_type=Prompt.Alert,
                info='Uploading file works with plan or case. Nitrate cannot '
                     'proceed with no plan or case ID.',
                next='javascript:window.history.go(-1);',
            )

        if request.FILES.get('upload_file'):
            upload_file = request.FILES['upload_file']

            try:
                upload_file.name.encode('utf8')
            except UnicodeEncodeError:
                return Prompt.render(
                    request=request,
                    info_type=Prompt.Alert,
                    info='Upload File name is not legal.',
                    next='javascript:window.history.go(-1);',
                )

            now = datetime.now()

            stored_name = f'{request.user.username}-{now}-{upload_file.name}'

            stored_file_name = os.path.join(
                settings.FILE_UPLOAD_DIR, stored_name).replace('\\', '/')
            stored_file_name = smart_str(stored_file_name)

            if upload_file.size > settings.MAX_UPLOAD_SIZE:
                return Prompt.render(
                    request=request,
                    info_type=Prompt.Alert,
                    info=f'You upload entity is too large. Please ensure the file is'
                         f' less than {settings.MAX_UPLOAD_SIZE} bytes.',
                    next='javascript:window.history.go(-1);',
                )

            # Create the upload directory when it's not exist
            if not os.path.exists(settings.FILE_UPLOAD_DIR):
                os.mkdir(settings.FILE_UPLOAD_DIR)

            if os.path.exists(stored_file_name):
                return Prompt.render(
                    request=request,
                    info_type=Prompt.Alert,
                    info=f"File named '{upload_file.name}' already exists in "
                         f"upload folder, please rename to another name for "
                         f"solve conflict.",
                    next='javascript:window.history.go(-1);',
                )

            with open(stored_file_name, 'wb+') as f:
                for chunk in upload_file.chunks():
                    f.write(chunk)

            # Write the file to database
            # store_file = open(upload_file_name, 'ro')
            ta = TestAttachment.objects.create(
                submitter_id=request.user.id,
                description=request.POST.get('description', None),
                file_name=upload_file.name,
                stored_name=stored_name,
                create_date=now,
                mime_type=upload_file.content_type
            )

            if request.POST.get('to_plan_id'):
                TestPlanAttachment.objects.create(
                    plan_id=int(to_plan_id),
                    attachment_id=ta.attachment_id,
                )
                return HttpResponseRedirect(
                    reverse('plan-attachment', args=[to_plan_id]))

            if request.POST.get('to_case_id'):
                TestCaseAttachment.objects.create(
                    attachment_id=ta.attachment_id,
                    case_id=int(to_case_id))
                return HttpResponseRedirect(
                    reverse('case-attachment', args=[to_case_id]))
        else:
            if 'to_plan_id' in request.POST:
                return HttpResponseRedirect(
                    reverse('plan-attachment', args=[to_plan_id]))
            if 'to_case_id' in request.POST:
                return HttpResponseRedirect(
                    reverse('case-attachment', args=[to_case_id]))


@require_GET
def check_file(request, file_id):
    """Download attachment file"""
    attachment = get_object_or_404(TestAttachment, pk=file_id)

    attachment_data = TestAttachmentData.objects.filter(
        attachment__attachment_id=file_id
    ).first()

    # First try to read file content from database.
    if attachment_data:
        # File content is not written into TestAttachmentData in upload_file,
        # this code path is dead now. Think about if file content should be
        # written into database in the future.
        contents = attachment_data.contents
    else:
        # File was not written into database, read it from configured file
        # system.
        stored_file_name = os.path.join(
            settings.FILE_UPLOAD_DIR,
            unquote(attachment.stored_name or attachment.file_name)
        ).replace('\\', '/')

        if not os.path.exists(stored_file_name):
            raise Http404(f'Attachment file {stored_file_name} does not exist.')

        try:
            with open(stored_file_name, 'rb') as f:
                contents = f.read()
        except IOError:
            msg = 'Cannot read attachment file from server.'
            log.exception(msg)
            return Prompt.render(
                request=request,
                info_type=Prompt.Alert,
                info=msg,
                next='javascript:window.history.go(-1);',
            )

    response = HttpResponse(contents, content_type=str(attachment.mime_type))
    file_name = smart_str(attachment.file_name)
    response['Content-Disposition'] = f'attachment; filename="{file_name}"'
    return response


def able_to_delete_attachment(request, file_id):
    """
    These are allowed to delete attachment -
        1. super user
        2. attachments's submitter
        3. testplan's author or owner
        4. testcase's owner
    """

    user = request.user
    if user.is_superuser:
        return True

    attach = TestAttachment.objects.get(attachment_id=file_id)
    if user.pk == attach.submitter_id:
        return True

    if 'from_plan' in request.GET:
        plan_id = int(request.GET['from_plan'])
        plan = TestPlan.objects.get(plan_id=plan_id)
        return user.pk == plan.owner_id or user.pk == plan.author_id

    if 'from_case' in request.GET:
        case_id = int(request.GET['from_case'])
        case = TestCase.objects.get(case_id=case_id)
        return user.pk == case.author_id

    return False


# Delete Attachment
def delete_file(request, file_id):
    ajax_response = {'rc': 0, 'response': 'ok'}
    DELEFAILURE = 1
    AUTHUNSUCCESS = 2

    state = able_to_delete_attachment(request, file_id)
    if not state:
        ajax_response['rc'] = AUTHUNSUCCESS
        ajax_response['response'] = 'auth_failure'
        return JsonResponse(ajax_response)

    # Delete plan's attachment
    if 'from_plan' in request.GET:
        try:
            plan_id = int(request.GET['from_plan'])
            attachment = TestPlanAttachment.objects.filter(attachment=file_id,
                                                           plan_id=plan_id)
            attachment.delete()
        except TestPlanAttachment.DoesNotExist:
            ajax_response['rc'] = DELEFAILURE
            ajax_response['response'] = 'failure'
        return JsonResponse(ajax_response)

    # Delete cases' attachment
    elif 'from_case' in request.GET:
        try:
            case_id = int(request.GET['from_case'])
            attachment = TestCaseAttachment.objects.filter(attachment=file_id,
                                                           case_id=case_id)
            attachment.delete()
        except TestCaseAttachment.DoesNotExist:
            ajax_response['rc'] = DELEFAILURE
            ajax_response['response'] = 'failure'

        return JsonResponse(ajax_response)



# -*- coding: utf-8 -*-

import json
import re

from http import HTTPStatus
from urllib.parse import urlparse, parse_qs

from django import test
from django.contrib.auth.models import Permission
from django.contrib.auth.models import User
from django.core.exceptions import ValidationError

from tcms.testcases.models import TestCaseStatus
from tcms.testruns.models import TestCaseRunStatus
from tests import factories as f

__all__ = (
    'user_should_have_perm',
    'remove_perm_from_user',
    'BasePlanCase',
    'BaseCaseRun',
    'encode',
    'HelperAssertions',
)


def encode(s):
    """For running tests in Python 3

    This is added for running tests successfully with Python 3 while porting to
    Python 3 compatibility. But, it is really necessary to think about the
    XMLPRC API that strings returned should be in bytestring or not.
    """
    if isinstance(s, str):
        return s.encode()
    else:
        return s


def user_should_have_perm(user, perm):
    if isinstance(perm, str):
        try:
            app_label, codename = perm.split('.')
        except ValueError:
            raise ValueError('%s is not valid. Should be in format app_label.perm_codename')
        else:
            if not app_label or not codename:
                raise ValueError('Invalid app_label or codename')
            get_permission = Permission.objects.get
            user.user_permissions.add(
                get_permission(content_type__app_label=app_label, codename=codename))
    elif isinstance(perm, Permission):
        user.user_permissions.add(perm)
    else:
        raise TypeError('perm should be an instance of either basestring or Permission')


def remove_perm_from_user(user, perm):
    """Remove a permission from an user"""

    if isinstance(perm, str):
        try:
            app_label, codename = perm.split('.')
        except ValueError:
            raise ValueError('%s is not valid. Should be in format app_label.perm_codename')
        else:
            if not app_label or not codename:
                raise ValueError('Invalid app_label or codename')
            get_permission = Permission.objects.get
            user.user_permissions.remove(
                get_permission(content_type__app_label=app_label, codename=codename))
    elif isinstance(perm, Permission):
        user.user_permissions.remove(perm)
    else:
        raise TypeError('perm should be an instance of either basestring or Permission')


def create_request_user(username=None, password=None):
    if username:
        user = f.UserFactory(username=username)
    else:
        user = f.UserFactory()
    if password:
        user.set_password(password)
    else:
        user.set_password('password')
    user.save()
    return user


class HelperAssertions:
    """Helper assertion methods"""

    def assert200(self, response):
        self.assertEqual(HTTPStatus.OK, response.status_code)

    def assert301(self, response):
        self.assertEqual(HTTPStatus.MOVED_PERMANENTLY, response.status_code)

    def assert302(self, response):
        self.assertEqual(HTTPStatus.FOUND, response.status_code)

    def assert400(self, response):
        self.assertEqual(HTTPStatus.BAD_REQUEST, response.status_code)

    def assert403(self, response):
        self.assertEqual(HTTPStatus.FORBIDDEN, response.status_code)

    def assert404(self, response):
        self.assertEqual(HTTPStatus.NOT_FOUND, response.status_code)

    def assert500(self, response):
        self.assertEqual(
            HTTPStatus.INTERNAL_SERVER_ERROR, response.status_code)

    def assertJsonResponse(self, response, expected, status_code=200):
        self.assertEqual(status_code, response.status_code)
        self.assertEqual(expected, json.loads(response.content))

    def assert_url(self, expected_url, url):
        """Check if two URL are same

        Assertions are called inside this this method. If anything is
        different, it will fail immediately.

        :param str expected_url: expected URL compare.
        :param str url: the URL to check if it is same as the expected URL.
        """
        url = urlparse(url)
        expected_url = urlparse(expected_url)

        self.assertEqual(expected_url.scheme, url.scheme)
        self.assertEqual(expected_url.netloc, url.netloc)
        self.assertEqual(expected_url.path, url.path)
        self.assertEqual(parse_qs(expected_url.query), parse_qs(url.query))

    def assertValidationError(self, field, message_regex, func, *args, **kwargs):
        """Assert django.core.exceptions.ValidationError is raised with expected message"""
        try:
            func(*args, **kwargs)
        except Exception as e:
            self.assertIsInstance(
                e, ValidationError, f'Exception {e} is not a ValidationError.')
            self.assertIn(field, e.message_dict,
                          f'Field {field} is not included in errors.')
            matches = [re.search(message_regex, item) is not None
                       for item in e.message_dict[field]]
            self.assertTrue(any(matches), 'Expected match message is not included.')
        else:
            self.fail('ValidationError is not raised.')


class AuthMixin(object):

    @classmethod
    def setUpTestData(cls):
        cls.tester = User.objects.create_user(
            username='nitrate-tester',
            email='nitrate-tester@example.com')
        cls.tester.set_password('password')
        cls.tester.save()

    def login_tester(self, user=None, password=None):
        """Login tester user for test

        Login pre-created tester user by default. If both user and password
        are given, login that user instead.
        """
        if user and password:
            login_user = user
            login_password = password
        else:
            login_user = self.tester
            login_password = 'password'

        self.client.login(username=login_user.username,
                          password=login_password)


class NitrateTestCase(test.TestCase):
    """Base test case for writing tests for Nitrate"""

    @classmethod
    def create_bz_tracker(cls):
        """Helper function to create a Bugzilla issue tracker"""
        return f.IssueTrackerFactory(
            name='bz',
            service_url='http://bugs.example.com/',
            issue_report_endpoint='/enter_bug.cgi',
            issue_url_fmt='http://bugs.example.com/?id={issue_key}',
            issues_display_url_fmt='http://bugs.example.com/?bug_id={issue_keys}',
            validate_regex=r'^\d+$')

    @classmethod
    def create_jira_tracker(cls):
        """Helper function to create a Bugzilla issue tracker"""
        return f.IssueTrackerFactory(
            name='jira',
            service_url='http://jira.example.com/',
            issue_report_endpoint='/enter_bug.cgi',
            issue_url_fmt='http://jira.example.com/browse/{issue_key}',
            issues_display_url_fmt='http://jira.example.com/?jql=issuekey in ({issue_keys})',
            validate_regex=r'^[A-Z]+-\d+$')


class BasePlanCase(HelperAssertions, AuthMixin, NitrateTestCase):
    """Base test case by providing essential Plan and Case objects used in tests"""

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()

        cls.case_status_confirmed = TestCaseStatus.objects.get(name='CONFIRMED')
        cls.case_status_proposed = TestCaseStatus.objects.get(name='PROPOSED')

        cls.product = f.ProductFactory(name='Nitrate')
        cls.version = f.VersionFactory(value='0.1', product=cls.product)

        cls.plan = f.TestPlanFactory(
            author=cls.tester,
            owner=cls.tester,
            product=cls.product,
            product_version=cls.version)

        cls.case = f.TestCaseFactory(
            summary='Test case 0',
            author=cls.tester,
            default_tester=None,
            reviewer=cls.tester,
            case_status=cls.case_status_confirmed,
            plan=[cls.plan])
        cls.case_1 = f.TestCaseFactory(
            summary='Test case 1',
            author=cls.tester,
            default_tester=None,
            reviewer=cls.tester,
            case_status=cls.case_status_confirmed,
            plan=[cls.plan])
        cls.case_2 = f.TestCaseFactory(
            summary='Test case 2',
            author=cls.tester,
            default_tester=None,
            reviewer=cls.tester,
            case_status=cls.case_status_confirmed,
            plan=[cls.plan])
        cls.case_3 = f.TestCaseFactory(
            summary='Test case 3',
            author=cls.tester,
            default_tester=None,
            reviewer=cls.tester,
            case_status=cls.case_status_confirmed,
            plan=[cls.plan])

        cls.case_4 = f.TestCaseFactory(
            summary='Test case 4',
            author=cls.tester,
            default_tester=None,
            reviewer=cls.tester,
            case_status=cls.case_status_confirmed,
            plan=[cls.plan])
        cls.case_5 = f.TestCaseFactory(
            summary='Test case 5',
            author=cls.tester,
            default_tester=None,
            reviewer=cls.tester,
            case_status=cls.case_status_confirmed,
            plan=[cls.plan])
        cls.case_6 = f.TestCaseFactory(
            summary='Test case 6',
            author=cls.tester,
            default_tester=None,
            reviewer=cls.tester,
            case_status=cls.case_status_confirmed,
            plan=[cls.plan])


class BaseCaseRun(BasePlanCase):
    """Base test case containing test run and case runs"""

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()

        cls.case_run_status_idle = TestCaseRunStatus.objects.get(name='IDLE')

        cls.build = f.TestBuildFactory(product=cls.product)

        cls.test_run = f.TestRunFactory(
            product_version=cls.version,
            plan=cls.plan,
            build=cls.build,
            manager=cls.tester,
            default_tester=cls.tester)

        cls.case_run_1, cls.case_run_2, cls.case_run_3 = [
            f.TestCaseRunFactory(
                assignee=cls.tester,
                tested_by=cls.tester,
                run=cls.test_run,
                build=cls.build,
                case_run_status=cls.case_run_status_idle,
                case=case,
                sortkey=i * 10)
            for i, case in enumerate((cls.case_1, cls.case_2, cls.case_3), 1)
        ]

        cls.test_run_1 = f.TestRunFactory(
            product_version=cls.version,
            plan=cls.plan,
            build=cls.build,
            manager=cls.tester,
            default_tester=cls.tester)

        cls.case_run_4, cls.case_run_5, cls.case_run_6 = [
            f.TestCaseRunFactory(
                assignee=cls.tester,
                tested_by=cls.tester,
                run=cls.test_run_1,
                build=cls.build,
                case_run_status=cls.case_run_status_idle,
                case=case,
                sortkey=i * 10)
            for i, case in enumerate((cls.case_4, cls.case_5, cls.case_6), 1)
        ]



# -*- coding: utf-8 -*-
from django.conf.urls import url

from .. import views

urlpatterns = [
    url(r'^$', views.all, name='runs-all'),
    url(r'^ajax/$', views.ajax_search, name='runs-ajax-search'),
    url(r'^env_value/$', views.env_value, name='runs-env-value'),
    url(r'^clone/$', views.clone, name='runs-clone'),
]



# -*- coding: utf-8 -*-

import datetime
import functools
import itertools
import json
import urllib

from operator import add, itemgetter

from django.utils.decorators import method_decorator
from django.conf import settings
from django.contrib.auth.decorators import permission_required
from django.contrib.auth.mixins import PermissionRequiredMixin
from django.core.exceptions import ObjectDoesNotExist
from django.urls import reverse
from django.db.models import Count
from django.db.models import Q
from django.http import Http404, HttpResponsePermanentRedirect
from django.http import HttpResponse, HttpResponseRedirect
from django.http import HttpResponseBadRequest, JsonResponse
from django.shortcuts import get_object_or_404, render
from django.template.loader import get_template
from django.views.decorators.http import require_GET
from django.views.decorators.http import require_http_methods
from django.views.generic import View
from uuslug import slugify

from tcms.core.db import SQLExecution
from tcms.core.models import TCMSLog
from tcms.core.raw_sql import RawSQL
from tcms.core.utils import checksum
from tcms.core.utils import DataTableResult
from tcms.core.views import Prompt
from tcms.management.models import TCMSEnvGroup, Component
from tcms.search.views import remove_from_request_path
from tcms.search.order import order_plan_queryset
from tcms.testcases.data import get_exported_cases_and_related_data
from tcms.testcases.forms import SearchCaseForm, QuickSearchCaseForm
from tcms.testcases.models import TestCaseStatus
from tcms.testcases.models import TestCase, TestCasePlan
from tcms.testcases.views import get_selected_testcases
from tcms.testplans.forms import ClonePlanForm
from tcms.testplans.forms import EditPlanForm
from tcms.testplans.forms import ImportCasesViaXMLForm
from tcms.testplans.forms import NewPlanForm
from tcms.testplans.forms import PlanComponentForm
from tcms.testplans.forms import SearchPlanForm
from tcms.testplans import sqls
from tcms.testplans.models import TestPlan, TestPlanComponent
from tcms.testplans.models import TestPlanEmailSettings
from tcms.testruns.models import TestRun, TestCaseRun


MODULE_NAME = "testplans"


# _____________________________________________________________________________
# helper functons


def update_plan_email_settings(tp, form):
    """Update testplan's email settings"""
    tp.email_settings.notify_on_plan_update = form.cleaned_data[
        'notify_on_plan_update']
    tp.email_settings.notify_on_plan_delete = form.cleaned_data[
        'notify_on_plan_delete']
    tp.email_settings.notify_on_case_update = form.cleaned_data[
        'notify_on_case_update']
    tp.email_settings.auto_to_plan_owner = form.cleaned_data['auto_to_plan_owner']
    tp.email_settings.auto_to_plan_author = form.cleaned_data['auto_to_plan_author']
    tp.email_settings.auto_to_case_owner = form.cleaned_data['auto_to_case_owner']
    tp.email_settings.auto_to_case_default_tester = form.cleaned_data[
        'auto_to_case_default_tester']
    tp.email_settings.save()


# _____________________________________________________________________________
# view functons


@require_http_methods(['GET', 'POST'])
@permission_required('testplans.add_testplan')
def new(request, template_name='plan/new.html'):
    """New testplan"""

    SUB_MODULE_NAME = "new_plan"

    # If the form has been submitted...
    if request.method == 'POST':
        # A form bound to the POST data
        form = NewPlanForm(request.POST, request.FILES)
        form.populate(product_id=request.POST.get('product'))

        # Process the upload plan document
        if form.is_valid():
            if form.cleaned_data.get('upload_plan_text'):
                # Set the summary form field to the uploaded text
                form.data['text'] = form.cleaned_data['text']

                # Generate the form
                context_data = {
                    'module': MODULE_NAME,
                    'sub_module': SUB_MODULE_NAME,
                    'form': form,
                }
                return render(request, template_name, context=context_data)

        # Process the test plan submit to the form

        if form.is_valid():
            tp = TestPlan.objects.create(
                product=form.cleaned_data['product'],
                author=request.user,
                owner=request.user,
                product_version=form.cleaned_data['product_version'],
                type=form.cleaned_data['type'],
                name=form.cleaned_data['name'],
                create_date=datetime.datetime.now(),
                extra_link=form.cleaned_data['extra_link'],
                parent=form.cleaned_data['parent'],
            )

            TestPlanEmailSettings.objects.create(plan=tp)

            # Add test plan text
            if request.user.has_perm('testplans.add_testplantext'):
                tp.add_text(author=request.user, plan_text=form.cleaned_data['text'])

            # Add test plan environment groups
            if request.user.has_perm('management.add_tcmsenvplanmap'):
                if request.POST.get('env_group'):
                    env_groups = TCMSEnvGroup.objects.filter(
                        id__in=request.POST.getlist('env_group')
                    )

                    for env_group in env_groups:
                        tp.add_env_group(env_group=env_group)

            return HttpResponseRedirect(
                reverse('plan-get', args=[tp.plan_id])
            )
    else:
        form = NewPlanForm()

    context_data = {
        'module': MODULE_NAME,
        'sub_module': SUB_MODULE_NAME,
        'form': form,
    }
    return render(request, template_name, context=context_data)


@require_GET
@permission_required('testplans.delete_testplan')
def delete(request, plan_id):
    """Delete testplan"""
    if request.GET.get('sure', 'no') == 'no':
        # TODO: rewrite the response
        plan_delete_url = reverse('plan-delete', args=[plan_id])
        return HttpResponse(
            "<script>"
            "if (confirm('Are you sure you want to delete this plan %s?\\n\\n"
            "Click OK to delete or cancel to come back'))"
            "{ window.location.href='%s?sure=yes' }"
            "else { history.go(-1) }"
            "</script>" % (plan_id, plan_delete_url)
        )
    elif request.GET.get('sure') == 'yes':
        tp = get_object_or_404(TestPlan, plan_id=plan_id)

        try:
            tp.delete()
            return HttpResponse(
                "<script>window.location.href='%s'</script>" % reverse(
                    'tcms.testplans.views.all')
            )
        except Exception:
            return Prompt.render(
                request=request,
                info_type=Prompt.Info,
                info='Delete failed.',
                next='javascript:window.history.go(-1)',
            )
    else:
        return Prompt.render(
            request=request,
            info_type=Prompt.Info,
            info='Nothing yet.',
            next='javascript:window.history.go(-1)',
        )


@require_GET
def all(request, template_name='plan/all.html'):
    """Display all testplans"""
    # Define the default sub module
    SUB_MODULE_NAME = 'plans'
    # TODO: this function now only performs a forward feature, no queries
    # need here. All of it will be removed in the future.
    # If it's not a search the page will be blank
    tps = TestPlan.objects.none()
    query_result = False
    order_by = request.GET.get('order_by', 'create_date')
    asc = bool(request.GET.get('asc', None))
    # if it's a search request the page will be fill
    if list(request.GET.items()):
        search_form = SearchPlanForm(request.GET)
        if request.GET.get('product'):
            search_form.populate(product_id=request.GET['product'])
        else:
            search_form.populate()

        if search_form.is_valid():
            # Detemine the query is the user's plans and change the sub
            # module value
            author = request.GET.get('author')
            if author and request.user.is_authenticated:
                if author == request.user.username or author == request.user.email:
                    SUB_MODULE_NAME = "my_plans"

            query_result = True
            # build a QuerySet:
            tps = TestPlan.list(search_form.cleaned_data)
            tps = tps.select_related('author', 'type', 'product')

            # We want to get the number of cases and runs, without doing
            # lots of per-test queries.
            #
            # Ideally we would get the case/run counts using m2m field tricks
            # in the ORM
            # Unfortunately, Django's select_related only works on ForeignKey
            # relationships, not on ManyToManyField attributes
            # See http://code.djangoproject.com/ticket/6432

            # SQLAlchemy can handle this kind of thing in several ways.
            # Unfortunately we're using Django

            # The cleanest way I can find to get it into one query is to
            # use QuerySet.extra()
            # See http://docs.djangoproject.com/en/dev/ref/models/querysets
            tps = tps.extra(select={
                'num_cases': RawSQL.num_cases,
                'num_runs': RawSQL.num_runs,
                'num_children': RawSQL.num_plans,
            })
            tps = order_plan_queryset(tps, order_by, asc)
    else:
        # Set search active plans only by default
        # I wish to use 'default' argument, as the same as in ModelForm
        # But it does not seem to work
        search_form = SearchPlanForm(initial={'is_active': True})

    if request.GET.get('action') == 'clone_case':
        template_name = 'case/clone_select_plan.html'
        tps = tps.order_by('name')

    if request.GET.get('t') == 'ajax':
        data = [
            dict(
                pk=plan.pk,
                name=plan.name,
                is_active=plan.is_active,
                parent_id=plan.parent_id,
                num_cases=plan.num_cases,
                num_runs=plan.num_runs,
                num_children=plan.num_children,
                get_url_path=plan.get_absolute_url(),
            )
            for plan in tps
        ]
        return JsonResponse(data, safe=False)

    if request.GET.get('t') == 'html':
        if request.GET.get('f') == 'preview':
            template_name = 'plan/preview.html'

    query_url = remove_from_request_path(request, 'order_by')
    if asc:
        query_url = remove_from_request_path(query_url, 'asc')
    else:
        query_url = '%s&asc=True' % query_url
    page_type = request.GET.get('page_type', 'pagination')
    query_url_page_type = remove_from_request_path(request, 'page_type')
    if query_url_page_type:
        query_url_page_type = remove_from_request_path(query_url_page_type, 'page')

    context_data = {
        'module': MODULE_NAME,
        'sub_module': SUB_MODULE_NAME,
        'test_plans': tps,
        'query_result': query_result,
        'search_plan_form': search_form,
        'query_url': query_url,
        'query_url_page_type': query_url_page_type,
        'page_type': page_type
    }
    return render(request, template_name, context=context_data)


def get_number_of_plans_cases(plan_ids):
    """Get the number of cases related to each plan

    Arguments:
    - plan_ids: a tuple or list of TestPlans' id

    Return value:
    Return value is an dict object, where key is plan_id and the value is the
    total count.
    """
    qs = TestCasePlan.objects.filter(plan__in=plan_ids)
    qs = qs.values('plan').annotate(
        total_count=Count('pk')).order_by('-plan')
    return {item['plan']: item['total_count'] for item in qs}


def get_number_of_plans_runs(plan_ids):
    """Get the number of runs related to each plan

    Arguments:
    - plan_ids: a tuple or list of TestPlans' id

    Return value:
    Return value is an dict object, where key is plan_id and the value is the
    total count.
    """
    qs = TestRun.objects.filter(plan__in=plan_ids)
    qs = qs.values('plan').annotate(
        total_count=Count('pk')).order_by('-plan')
    return {item['plan']: item['total_count'] for item in qs}


def get_number_of_children_plans(plan_ids):
    """Get the number of children plans related to each plan

    Arguments:
    - plan_ids: a tuple or list of TestPlans' id

    Return value:
    Return value is an dict object, where key is plan_id and the value is the
    total count.
    """
    qs = TestPlan.objects.filter(parent__in=plan_ids)
    qs = qs.values('parent').annotate(
        total_count=Count('parent')).order_by('-parent')
    return {item['parent']: item['total_count'] for item in qs}


def calculate_stats_for_testplans(plans):
    """Attach the number of cases and runs for each TestPlan

    Arguments:
    - plans: the queryset of TestPlans

    Return value:
    A list of TestPlans, each of which is attached the statistics which is
    with prefix cal meaning calculation result.
    """
    plan_ids = [plan.pk for plan in plans]
    cases_counts = get_number_of_plans_cases(plan_ids)
    runs_counts = get_number_of_plans_runs(plan_ids)
    children_counts = get_number_of_children_plans(plan_ids)

    # Attach calculated statistics to each object of TestPlan
    for plan in plans:
        setattr(plan, 'cal_cases_count', cases_counts.get(plan.pk, 0))
        setattr(plan, 'cal_runs_count', runs_counts.get(plan.pk, 0))
        setattr(plan, 'cal_children_count', children_counts.get(plan.pk, 0))

    return plans


@require_GET
def ajax_search(request, template_name='plan/common/json_plans.txt'):
    """Display all testplans"""
    # Define the default sub module

    # If it's not a search the page will be blank
    tps = TestPlan.objects.none()
    # if it's a search request the page will be fill
    if list(request.GET.items()):
        search_form = SearchPlanForm(request.GET)
        if request.GET.get('product'):
            search_form.populate(product_id=request.GET['product'])
        else:
            search_form.populate()
        if search_form.is_valid():
            # Detemine the query is the user's plans and change the sub
            # module value
            author = request.GET.get('author__email__startswith')
            if author and len(search_form.changed_data) == 1:
                if request.user.is_authenticated:
                    if author == request.user.username or author == request.user.email:
                        tps = TestPlan.objects.filter(
                            Q(author__email__startswith=author) |
                            Q(owner__email__startswith=author)
                        ).distinct()
            else:
                tps = TestPlan.list(search_form.cleaned_data)
                tps = tps.select_related('author', 'owner', 'type', 'product')

    # columnIndexNameMap is required for correct sorting behavior, 5 should
    # be product, but we use run.build.product
    column_names = [
        '',
        'plan_id',
        'name',
        'author__username',
        'owner__username',
        'product',
        'product_version',
        'type',
        'num_cases',
        'num_runs',
        ''
    ]
    return ajax_response(request, tps, column_names,
                         'plan/common/json_plans.txt')


def ajax_response(request, queryset, column_names, template_name):
    """json template for the ajax request for searching"""
    dt = DataTableResult(request.GET, queryset, column_names)

    data = dt.get_response_data()
    data['querySet'] = calculate_stats_for_testplans(data['querySet'])

    # prepare the JSON with the response, consider using :
    # from django.template.defaultfilters import escapejs
    t = get_template(template_name)
    resp_data = t.render(data, request)
    return JsonResponse(json.loads(resp_data))


def get(request, plan_id, slug=None, template_name='plan/get.html'):
    """Display the plan details."""
    SUB_MODULE_NAME = 'plans'

    try:
        tp = TestPlan.objects.select_related().get(plan_id=plan_id)
        tp.latest_text = tp.latest_text()
    except ObjectDoesNotExist:
        raise Http404

    # redirect if has a cheated slug
    if slug != slugify(tp.name):
        return HttpResponsePermanentRedirect(tp.get_absolute_url())

    # Initial the case counter
    confirm_status_name = 'CONFIRMED'
    tp.run_case = tp.case.filter(case_status__name=confirm_status_name)
    tp.review_case = tp.case.exclude(case_status__name=confirm_status_name)

    context_data = {
        'module': MODULE_NAME,
        'sub_module': SUB_MODULE_NAME,
        'test_plan': tp,
        'xml_form': ImportCasesViaXMLForm(),
    }
    return render(request, template_name, context=context_data)


class AddCasesToRunsView(PermissionRequiredMixin, View):
    """View of adding cases to runs"""

    SUB_MODULE_NAME = 'runs'
    permission_required = 'testruns.change_testrun'
    template_name = 'plan/choose_testrun.html'

    def get(self, request, plan_id):
        plan = TestPlan.objects.filter(
            pk=int(plan_id)).defer('product_version').first()
        if plan is None:
            raise Http404

        # TODO: replace with plan.run.values(...)
        runs = TestRun.objects.filter(plan=plan_id).values(
            'pk', 'summary', 'build__name', 'manager__username')

        cases = get_selected_testcases(request).values(
            'pk', 'summary', 'author__username', 'create_date',
            'category__name', 'priority__value'
        )

        return render(request, self.template_name, context={
            'module': MODULE_NAME,
            'sub_module': self.SUB_MODULE_NAME,
            'plan_id': plan_id,
            'plan': plan,
            'test_runs': runs.iterator(),
            'test_cases': cases,
        })

    def post(self, request, plan_id):
        choosed_testrun_ids = request.POST.getlist('testrun_ids')
        to_be_added_cases = TestCase.objects.filter(
            pk__in=request.POST.getlist('case_ids'))

        plan_url = reverse('plan-get', args=[plan_id])

        # cases and runs are required in this process
        if not len(choosed_testrun_ids) or not len(to_be_added_cases):
            return Prompt.render(
                request=request,
                info_type=Prompt.Info,
                info='At least one test run and one case is required to add '
                     'cases to runs.',
                next=plan_url,
            )

        # Adding cases to runs by recursion
        for tr_id in choosed_testrun_ids:
            testrun = get_object_or_404(TestRun, run_id=tr_id)
            cases = TestCaseRun.objects.filter(run=tr_id)
            exist_cases_id = cases.values_list('case', flat=True)

            for testcase in to_be_added_cases:
                if testcase.case_id not in exist_cases_id:
                    testrun.add_case_run(case=testcase)

            estimated_time = functools.reduce(
                add,
                [nc.estimated_time for nc in to_be_added_cases])
            testrun.estimated_time = testrun.estimated_time + estimated_time
            testrun.save()

        return HttpResponseRedirect(plan_url)


@require_http_methods(['GET', 'POST'])
@permission_required('testplans.change_testplan')
def edit(request, plan_id, template_name='plan/edit.html'):
    """Edit test plan view"""
    # Define the default sub module
    SUB_MODULE_NAME = 'plans'

    try:
        tp = TestPlan.objects.select_related().get(plan_id=plan_id)
    except ObjectDoesNotExist:
        raise Http404

    # If the form is submitted
    if request.method == "POST":
        form = EditPlanForm(request.POST, request.FILES)
        if request.POST.get('product'):
            form.populate(product_id=request.POST['product'])
        else:
            form.populate()

        # FIXME: Error handle
        if form.is_valid():
            if form.cleaned_data.get('upload_plan_text'):
                # Set the summary form field to the uploaded text
                form.data['text'] = form.cleaned_data['text']

                # Generate the form
                context_data = {
                    'module': MODULE_NAME,
                    'sub_module': SUB_MODULE_NAME,
                    'form': form,
                    'test_plan': tp,
                }
                return render(request, template_name, context=context_data)

            if request.user.has_perm('testplans.change_testplan'):
                tp.name = form.cleaned_data['name']
                tp.parent = form.cleaned_data['parent']
                tp.product = form.cleaned_data['product']
                tp.product_version = form.cleaned_data['product_version']
                tp.type = form.cleaned_data['type']
                tp.is_active = form.cleaned_data['is_active']
                tp.extra_link = form.cleaned_data['extra_link']
                tp.owner = form.cleaned_data['owner']
                # IMPORTANT! tp.current_user is an instance attribute,
                # added so that in post_save, current logged-in user info
                # can be accessed.
                # Instance attribute is usually not a desirable solution.
                tp.current_user = request.user
                tp.save()

            if request.user.has_perm('testplans.add_testplantext'):
                new_text = request.POST.get('text')
                text_checksum = checksum(new_text)

                if not tp.text_exist() or text_checksum != tp.text_checksum():
                    tp.add_text(author=request.user,
                                plan_text=request.POST.get('text'),
                                text_checksum=text_checksum)

            if request.user.has_perm('management.change_tcmsenvplanmap'):
                tp.clear_env_groups()

                if request.POST.get('env_group'):
                    env_groups = TCMSEnvGroup.objects.filter(
                        id__in=request.POST.getlist('env_group'))

                    for env_group in env_groups:
                        tp.add_env_group(env_group=env_group)
            # Update plan email settings
            update_plan_email_settings(tp, form)
            return HttpResponseRedirect(
                reverse('plan-get', args=[plan_id, slugify(tp.name)]))
    else:
        # Generate a blank form
        # Temporary use one environment group in this case
        if tp.env_group.all():
            for env_group in tp.env_group.all():
                env_group_id = env_group.id
                break
        else:
            env_group_id = None

        form = EditPlanForm(initial={
            'name': tp.name,
            'product': tp.product_id,
            'product_version': tp.product_version_id,
            'type': tp.type_id,
            'text': tp.latest_text() and tp.latest_text().plan_text or '',
            'parent': tp.parent_id,
            'env_group': env_group_id,
            'is_active': tp.is_active,
            'extra_link': tp.extra_link,
            'owner': tp.owner,
            'auto_to_plan_owner': tp.email_settings.auto_to_plan_owner,
            'auto_to_plan_author': tp.email_settings.auto_to_plan_author,
            'auto_to_case_owner': tp.email_settings.auto_to_case_owner,
            'auto_to_case_default_tester': tp.email_settings.auto_to_case_default_tester,
            'notify_on_plan_update': tp.email_settings.notify_on_plan_update,
            'notify_on_case_update': tp.email_settings.notify_on_case_update,
            'notify_on_plan_delete': tp.email_settings.notify_on_plan_delete,
        })
        form.populate(product_id=tp.product_id)

    context_data = {
        'module': MODULE_NAME,
        'sub_module': SUB_MODULE_NAME,
        'test_plan': tp,
        'form': form,
    }
    return render(request, template_name, context=context_data)


@require_http_methods(['GET', 'POST'])
@permission_required('testplans.add_testplan')
def clone(request, template_name='plan/clone.html'):
    """Clone testplan"""
    SUB_MODULE_NAME = 'plans'

    req_data = request.GET or request.POST
    if 'plan' not in req_data:
        return Prompt.render(
            request=request,
            info_type=Prompt.Info,
            info='At least one plan is required by clone function.',
            next='javascript:window.history.go(-1)',
        )

    tps = TestPlan.objects.filter(pk__in=req_data.getlist('plan'))

    if not tps:
        return Prompt.render(
            request=request,
            info_type=Prompt.Info,
            info='The plan you specify does not exist in database.',
            next='javascript:window.history.go(-1)',
        )

    # Clone the plan if the form is submitted
    if request.method == "POST":
        clone_form = ClonePlanForm(request.POST)
        clone_form.populate(product_id=request.POST.get('product_id'))

        if clone_form.is_valid():
            clone_options = clone_form.cleaned_data

            # Create new test plan.
            for tp in tps:

                new_name = len(tps) == 1 and clone_options['name'] or None

                clone_params = {
                    # Cloned plan properties
                    'new_name': new_name,
                    'product': clone_options['product'],
                    'version': clone_options['product_version'],
                    'set_parent': clone_options['set_parent'],

                    # Related data
                    'copy_texts': clone_options['copy_texts'],
                    'copy_attachments': clone_options['copy_attachements'],
                    'copy_environment_group': clone_options['copy_environment_group'],

                    # Link or copy cases
                    'link_cases': clone_options['link_testcases'],
                    'copy_cases': clone_options['copy_testcases'],
                    'default_component_initial_owner': request.user,
                }

                assign_me_as_plan_author = not clone_options['keep_orignal_author']
                if assign_me_as_plan_author:
                    clone_params['new_original_author'] = request.user

                assign_me_as_copied_case_author = (
                    clone_options['copy_testcases'] and
                    not clone_options['maintain_case_orignal_author']
                )
                if assign_me_as_copied_case_author:
                    clone_params['new_case_author'] = request.user

                assign_me_as_copied_case_default_tester = (
                    clone_options['copy_testcases'] and
                    not clone_options['keep_case_default_tester']
                )
                if assign_me_as_copied_case_default_tester:
                    clone_params['new_case_default_tester'] = request.user

                assign_me_as_text_author = not clone_options['copy_texts']
                if assign_me_as_text_author:
                    clone_params['default_text_author'] = request.user

                cloned_plan = tp.clone(**clone_params)

            if len(tps) == 1:
                return HttpResponseRedirect(
                    reverse('plan-get', args=[cloned_plan.plan_id]))
            else:
                args = {
                    'action': 'search',
                    'product': clone_form.cleaned_data['product'].id,
                    'product_version': clone_form.cleaned_data['product_version'].id,
                }
                url_args = urllib.parse.urlencode(args)
                return HttpResponseRedirect(
                    '{}?{}'.format(reverse('plans-all'), url_args))
    else:
        # Generate the default values for the form
        if len(tps) == 1:
            clone_form = ClonePlanForm(initial={
                'product': tps[0].product_id,
                'product_version': tps[0].product_version_id,
                'set_parent': True,
                'copy_texts': True,
                'copy_attachements': True,
                'copy_environment_group': True,
                'link_testcases': True,
                'copy_testcases': False,
                'maintain_case_orignal_author': True,
                'keep_case_default_tester': False,
                'name': tps[0].make_cloned_name(),
            })
            clone_form.populate(product_id=tps[0].product.id)
        else:
            clone_form = ClonePlanForm(initial={
                'set_parent': True,
                'copy_texts': True,
                'copy_attachements': True,
                'link_testcases': True,
                'copy_testcases': False,
                'maintain_case_orignal_author': True,
                'keep_case_default_tester': True,
            })

    context_data = {
        'module': MODULE_NAME,
        'sub_module': SUB_MODULE_NAME,
        'testplans': tps,
        'clone_form': clone_form,
    }
    return render(request, template_name, context=context_data)


def attachment(request, plan_id, template_name='plan/attachment.html'):
    """Manage attached files"""
    SUB_MODULE_NAME = 'plans'

    file_size_limit = settings.MAX_UPLOAD_SIZE
    limit_readable = int(file_size_limit) / 2 ** 20  # Mb

    tp = get_object_or_404(TestPlan, plan_id=plan_id)
    context_data = {
        'module': MODULE_NAME,
        'sub_module': SUB_MODULE_NAME,
        'test_plan': tp,
        'limit': file_size_limit,
        'limit_readable': str(limit_readable) + "Mb",
    }
    return render(request, template_name, context=context_data)


@require_GET
def text_history(request, plan_id, template_name='plan/history.html'):
    """View test plan text history"""
    SUB_MODULE_NAME = 'plans'

    tp = get_object_or_404(TestPlan, plan_id=int(plan_id))
    tptxts = tp.text.select_related('author').only('plan',
                                                   'create_date',
                                                   'plan_text',
                                                   'plan_text_version',
                                                   'author__email')
    selected_plan_text_version = int(request.GET.get('plan_text_version', 0))
    context_data = {
        'module': MODULE_NAME,
        'sub_module': SUB_MODULE_NAME,
        'testplan': tp,
        'test_plan_texts': tptxts,
        'select_plan_text_version': selected_plan_text_version,
    }
    return render(request, template_name, context=context_data)


class ReorderCasesView(View):
    """Reorder cases"""

    http_method_names = ['post']

    def post(self, request, plan_id):
        # Current we should rewrite all of cases belong to the plan.
        # Because the cases sortkey in database is chaos,
        # Most of them are None.

        if 'case' not in request.POST:
            return JsonResponse({
                'rc': 1,
                'response': 'At least one case is required to re-order.'
            })

        plan = get_object_or_404(TestPlan, pk=int(plan_id))

        case_ids = [int(id) for id in request.POST.getlist('case')]
        cases = TestCase.objects.filter(pk__in=case_ids).only('pk')

        for case in cases:
            new_sort_key = (case_ids.index(case.pk) + 1) * 10
            TestCasePlan.objects.filter(
                plan=plan, case=case).update(sortkey=new_sort_key)

        return JsonResponse({'rc': 0, 'response': 'ok'})


class LinkCasesView(View):
    """Link cases to plan"""

    permission_required = 'testcases.add_testcaseplan'

    def post(self, request, plan_id):
        plan = get_object_or_404(TestPlan.objects.only('pk'), pk=int(plan_id))
        case_ids = [int(id) for id in request.POST.getlist('case')]
        cases = TestCase.objects.filter(case_id__in=case_ids).only('pk')
        for case in cases:
            plan.add_case(case)
        return HttpResponseRedirect(reverse('plan-get', args=[plan_id]))


class LinkCasesSearchView(View):
    """Search cases for linking to plan"""

    template_name = 'plan/search_case.html'
    SUB_MODULE_NAME = 'plans'

    def get(self, request, plan_id):
        plan = get_object_or_404(TestPlan, pk=int(plan_id))

        normal_form = SearchCaseForm(initial={
            'product': plan.product_id,
            'product_version': plan.product_version_id,
            'case_status_id': TestCaseStatus.get('CONFIRMED')
        })
        quick_form = QuickSearchCaseForm()
        return render(self.request, self.template_name, {
            'module': MODULE_NAME,
            'sub_module': self.SUB_MODULE_NAME,
            'search_form': normal_form,
            'quick_form': quick_form,
            'test_plan': plan,
        })

    def post(self, request, plan_id):
        plan = get_object_or_404(TestPlan, pk=int(plan_id))

        search_mode = request.POST.get('search_mode')
        if search_mode == 'quick':
            form = quick_form = QuickSearchCaseForm(request.POST)
            normal_form = SearchCaseForm()
        else:
            form = normal_form = SearchCaseForm(request.POST)
            form.populate(product_id=request.POST.get('product'))
            quick_form = QuickSearchCaseForm()

        if form.is_valid():
            cases = TestCase.list(form.cleaned_data)
            cases = cases.select_related(
                'author', 'default_tester', 'case_status', 'priority'
            ).only(
                'pk', 'summary', 'create_date', 'author__email',
                'default_tester__email', 'case_status__name',
                'priority__value'
            ).exclude(
                case_id__in=plan.case.values_list('case_id', flat=True))

        context = {
            'module': MODULE_NAME,
            'sub_module': self.SUB_MODULE_NAME,
            'test_plan': plan,
            'test_cases': cases,
            'search_form': normal_form,
            'quick_form': quick_form,
            'search_mode': search_mode
        }
        return render(request, self.template_name, context=context)


class ImportCasesView(PermissionRequiredMixin, View):
    """Import cases to a plan"""

    permission_required = 'testcases.add_testcaseplan'

    def post(self, request, plan_id):
        plan = get_object_or_404(TestPlan.objects.only('pk'), pk=int(plan_id))
        xml_form = ImportCasesViaXMLForm(request.POST, request.FILES)
        if xml_form.is_valid():
            plan.import_cases(xml_form.cleaned_data['xml_file'])
            return HttpResponseRedirect(
                reverse('plan-get', args=[plan_id]) + '#testcases')
        else:
            return Prompt.render(
                request=request,
                info_type=Prompt.Alert,
                info=xml_form.errors,
                next=reverse('plan-get', args=[plan_id]) + '#testcases'
            )


class DeleteCasesView(View):
    """Delete selected cases from plan"""

    def post(self, request, plan_id):
        plan = get_object_or_404(TestPlan.objects.only('pk'), pk=int(plan_id))

        if 'case' not in request.POST:
            return JsonResponse({
                'rc': 1,
                'response': 'At least one case is required to delete.'
            })

        cases = get_selected_testcases(request).only('pk')

        # Log Action
        plan_log = TCMSLog(model=plan)
        for case in cases:
            plan_log.make(
                who=request.user,
                new_value=f'Remove case {case.pk} from plan {plan.pk}')
            case.log_action(
                who=request.user,
                new_value=f'Remove from plan {plan.pk}')
            plan.delete_case(case=case)

        return JsonResponse({'rc': 0, 'response': 'ok'})


class PlanComponentsActionView(View):
    """Manage a plan's components"""

    template_name = 'plan/get_component.html'

    def get(self, request):
        if 'plan' not in request.GET:
            return HttpResponseBadRequest('Plan ID is not in request.')
        plans = TestPlan.objects.filter(pk=int(request.GET['plan']))
        if not plans:
            return Http404('Plan ID {} does not exist.'.format(
                ', '.join(plans)))

        action = request.GET.get('a', 'get_component_list').lower()

        if action == 'get_form':
            return self.get_manage_form(request, plans)
        elif action == 'get_component_list':
            return self.get_default_component_list(request, plans[0])
        elif action == 'add':
            return self.add(request, plans[0], self._get_components())
        elif action == 'remove':
            components = self._get_components()
            return self.remove_components_from_plan(request, plans[0], components)
        elif action == 'update':
            return self.update_components(request, plans[0])

    def _get_components(self):
        if 'component' not in self.request.GET:
            return HttpResponseBadRequest('Component ID is not in request.')
        component_ids = [
            int(id) for id in self.request.GET.getlist('component')]
        return Component.objects.filter(pk__in=component_ids)

    @method_decorator(permission_required('testplans.add_testplancomponent'))
    def add(self, request, plan, components):
        """Add components to given plans"""
        list(map(plan.add_component, components))

    @method_decorator(permission_required('testplans.delete_testplancomponent'))
    def remove_components_from_plan(self, request, plan, components=None):
        """Remove existing components from plans

        :param plan: instance of TestPlan, from which to remove components
            from this plan.
        :param components: instances of Component, which will be removed.
        """
        if components is None:
            TestPlanComponent.objects.filter(plan=plan).delete()
        else:
            list(map(plan.remove_component, components))

        return self.get_default_component_list(request, plan)

    def update_components(self, request, plan):
        self.remove_components_from_plan(request, plan)
        self.add(request, plan, self._get_components())
        return self.get_default_component_list(request, plan)

    def get_manage_form(self, request, plans):
        """Return form content in order to select components"""
        plan_comps = TestPlanComponent.objects.filter(plan__in=plans)

        form = PlanComponentForm(tps=plans, initial={
            'component': plan_comps.values_list('component_id', flat=True),
        })

        q_format = request.GET.get('format', 'p')
        html = getattr(form, 'as_' + q_format)

        return HttpResponse(html())

    def get_default_component_list(self, request, plan):
        return render(request, self.template_name, context={'test_plan': plan})


def tree_view(request):
    """Whole tree view for plans"""
    # FIXME:


@require_GET
def printable(request, template_name='plan/printable.html'):
    """Create the printable copy for plan"""
    plan_pks = request.GET.getlist('plan')

    if not plan_pks:
        return Prompt.render(
            request=request,
            info_type=Prompt.Info,
            info='At least one target is required.')

    tps = TestPlan.objects.filter(pk__in=plan_pks).only('pk', 'name')

    def plan_generator():
        repeat = len(plan_pks)
        params_sql = ','.join(itertools.repeat('%s', repeat))
        sql = sqls.TP_PRINTABLE_CASE_TEXTS % (params_sql, params_sql)
        result_set = SQLExecution(sql, plan_pks * 2)
        group_data = itertools.groupby(result_set.rows, itemgetter('plan_id'))
        cases_dict = {key: list(values) for key, values in group_data}
        for tp in tps:
            tp.result_set = cases_dict.get(tp.plan_id, None)
            yield tp

    context_data = {
        'test_plans': plan_generator(),
    }

    return render(request, template_name, context=context_data)


@require_GET
def export(request, template_name='case/export.xml'):
    """Export the plan"""
    plan_pks = list(map(int, request.GET.getlist('plan')))

    if not plan_pks:
        return Prompt.render(
            request=request,
            info_type=Prompt.Info,
            info='At least one target is required.')

    context_data = {
        'cases_info': get_exported_cases_and_related_data(plan_pks),
    }

    timestamp = datetime.datetime.now()
    timestamp_str = '%02i-%02i-%02i' % (
        timestamp.year, timestamp.month, timestamp.day)

    response = render(request, template_name, context=context_data)
    filename = f'tcms-testcases-{timestamp_str}.xml'
    response['Content-Disposition'] = f'attachment; filename={filename}'
    return response



# -*- coding: utf-8 -*-

from django.conf.urls import url
from .. import views

urlpatterns = [
    url(r'^$', views.all, name='plans-all'),
    url(r'^new/$', views.new, name='plans-new'),
    url(r'^ajax/$', views.ajax_search, name='plans-ajax-search'),
    url(r'^treeview/$', views.tree_view, name='plans-treeview'),
    url(r'^clone/$', views.clone, name='plans-clone'),
    url(r'^printable/$', views.printable, name='plans-printable'),
    url(r'^export/$', views.export, name='plans-export'),
    # url(r'^component/$', views.component, name='plans-component'),

    url(r'^component/$', views.PlanComponentsActionView.as_view(),
        name='plans-component-actions'),
]



# -*- coding: utf-8 -*-
from django.conf.urls import url
from . import views

urlpatterns = [
    url(r'^post/$', views.post, name='comments-post'),
    url(r'^delete/$', views.DeleteCommentView.as_view(), name='comments-delete'),
]



# -*- coding: utf-8 -*-

from django.conf.urls import url

from .. import views
from tcms.testruns import views as testruns_views

urlpatterns = [
    url(r'^(?P<case_id>\d+)/$', views.get, name='case-get'),
    url(r'^(?P<case_id>\d+)/edit/$', views.edit, name='case-edit'),
    url(r'^(?P<case_id>\d+)/history/$', views.text_history, name='case-text-history'),
    url(r'^(?P<case_id>\d+)/attachment/$', views.ListCaseAttachmentsView.as_view(),
        name='case-attachment'),
    url(r'^(?P<case_id>\d+)/log/$', views.get_log, name='case-get-log'),
    url(r'^(?P<case_id>\d+)/issue/$', views.manage_case_issues, name='case-issue'),
    url(r'^(?P<case_id>\d+)/plan/$', views.plan, name='case-plan'),
    url(r'^(?P<case_id>\d+)/readonly-pane/$', views.SimpleTestCaseView.as_view(),
        name='case-readonly-pane'),
    url(r'^(?P<case_id>\d+)/review-pane/$', views.TestCaseReviewPaneView.as_view(),
        name='case-review-pane'),
    url(r'^(?P<case_id>\d+)/caserun-list-pane/$', views.TestCaseCaseRunListPaneView.as_view(),
        name='caserun-list-pane'),
    url(r'^(?P<case_id>\d+)/caserun-simple-pane/$', views.TestCaseSimpleCaseRunView.as_view(),
        name='caserun-simple-pane'),
    url(r'^(?P<case_id>\d+)/caserun-detail-pane/$', views.TestCaseCaseRunDetailPanelView.as_view(),
        name='caserun-detail-pane'),

    url(r'^(?P<plan_id>\d+)/runs/$', testruns_views.load_runs_of_one_plan,
        name='load_runs_of_one_plan_url'),
]



# -*- coding: utf-8 -*-
# FIXME: Use exception to replace the feature

from django.shortcuts import render


class Prompt:
    """Common dialog to prompt to users"""

    Alert = 'alert'
    Info = 'info'

    def __init__(self, request, info_type=None, info=None, next=None):
        super().__init__()
        self.request = request
        self.info_type = info_type
        self.info = info
        self.next = next

    @classmethod
    def render(cls, request, info_type=None, info=None, next=None):
        """Generate the html to response"""
        return render(request, 'prompt.html', context={
            'type': info_type,
            'info': info,
            'next': next
        })



# -*- coding: utf-8 -*-

from django.conf.urls import url

from .. import views
from tcms.testruns import views as testruns_views

urlpatterns = [
    url(r'^(?P<plan_id>\d+)/$', views.get, name='plan-get'),
    url(r'^(?P<plan_id>\d+)/(?P<slug>[-\w\d]+)$', views.get, name='plan-get'),
    url(r'^(?P<plan_id>\d+)/delete/$', views.delete, name='plan-delete'),

    url(r'^(?P<plan_id>\d+)/chooseruns/$', views.AddCasesToRunsView.as_view(),
        name='plan-choose-run'),

    url(r'^(?P<plan_id>\d+)/edit/$', views.edit, name='plan-edit'),
    url(r'^(?P<plan_id>\d+)/attachment/$', views.attachment, name='plan-attachment'),
    url(r'^(?P<plan_id>\d+)/history/$', views.text_history, name='plan-text-history'),

    url(r'^(?P<plan_id>\d+)/reorder-cases/$', views.ReorderCasesView.as_view(),
        name='plan-reorder-cases'),
    url(r'^(?P<plan_id>\d+)/link-cases/$', views.LinkCasesView.as_view(),
        name='plan-link-cases'),
    url(r'^(?P<plan_id>\d+)/link-cases/search/$', views.LinkCasesSearchView.as_view(),
        name='plan-search-cases-for-link'),
    url(r'^(?P<plan_id>\d+)/import-cases/$', views.ImportCasesView.as_view(),
        name='plan-import-cases'),
    url(r'^(?P<plan_id>\d+)/delete-cases/$', views.DeleteCasesView.as_view(),
        name='plan-delete-cases'),

    url(r'^(?P<plan_id>\d+)/runs/$', testruns_views.load_runs_of_one_plan,
        name='load_runs_of_one_plan_url'),
]



# -*- coding: utf-8 -*-
from django import template
from django.utils.safestring import mark_safe, SafeData
from django.template.defaultfilters import stringfilter

register = template.Library()


@register.filter
@stringfilter
def split_as_option(value, splitter='|', autoescape=None):
    if not isinstance(value, SafeData):
        value = mark_safe(value)
    value = value.split(splitter)
    result = ""

    for v in value:
        result += f'<option value="{v}">{v}</option>\n'

    return mark_safe(result)


split_as_option.is_safe = True
split_as_option.needs_autoescape = True


@register.filter
@stringfilter
def split_as_value(value, splitter='|', autoescape=None):
    if not isinstance(value, SafeData):
        value = mark_safe(value)
    value = value.split(splitter)
    result = ""

    for v in value:
        result += '<span class="value">%s</span>' % v

    return mark_safe(result)


split_as_value.is_safe = True
split_as_value.needs_autoescape = True


@register.filter
@stringfilter
def split_as_inputbox(value, splitter='|', autoescape=None):
    if not isinstance(value, SafeData):
        value = mark_safe(value)
    value = value.split(splitter)
    result = ""

    for v in value:
        result += (
            f'<input id="id_btn_{v}" type="text" name="value" value="{v}" />'
            f'<input type="button" value="Del" '
            f'onclick="$(\'id_btn_{v}\').remove(); this.remove();" />'
        )

    return mark_safe(result)


split_as_inputbox.is_safe = True
split_as_inputbox.needs_autoescape = True



# -*- coding: utf-8 -*-


class CsrfDisableMiddleware:

    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        # Do default behavior for this middleware, as here is not where it gets
        # work done.
        return self.get_response(request)

    def process_view(self, request, callback, callback_args, callback_kwargs):
        setattr(request, '_dont_enforce_csrf_checks', True)



# -*- coding: utf-8 -*-

import json
import os
import shutil
import tempfile

from datetime import datetime

from django.db.models import Max
from django.test import RequestFactory
from django.urls import reverse
from django.conf import settings
from mock import patch

from tcms.core.files import able_to_delete_attachment
from tcms.management.models import TestAttachment
from tcms.testcases.models import TestCaseAttachment
from tcms.testplans.models import TestPlanAttachment
from tests import factories as f
from tests import BasePlanCase
from tests import create_request_user
from tests import user_should_have_perm


class TestUploadFile(BasePlanCase):
    """Test view upload_file"""

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()

        cls.upload_file_url = reverse('upload-file')

        cls.password = 'password'
        cls.user = create_request_user(username='uploader', password=cls.password)
        user_should_have_perm(cls.user, 'management.add_testattachment')
        user_should_have_perm(cls.user, 'testcases.add_testcaseattachment')

    @classmethod
    def setUpClass(cls):
        super().setUpClass()

        cls.file_upload_dir = tempfile.mkdtemp(
            prefix=f'{cls.__name__}-upload-dir')

    @classmethod
    def tearDownClass(cls):
        shutil.rmtree(cls.file_upload_dir)

        super().tearDownClass()

    def setUp(self):
        super().setUp()

        fd, self.upload_filename = tempfile.mkstemp(
            prefix=f'{self.__class__.__name__}-upload-file.txt')
        os.write(fd, b'abc' * 100)
        os.close(fd)

    def tearDown(self):
        os.remove(self.upload_filename)
        super().tearDown()

    def test_no_file_is_posted(self):
        self.client.login(username=self.user.username, password=self.password)

        response = self.client.post(reverse('upload-file'),
                                    {'to_plan_id': self.plan.pk})
        self.assertRedirects(
            response,
            reverse('plan-attachment', args=[self.plan.pk]))

        response = self.client.post(reverse('upload-file'),
                                    {'to_case_id': self.case_1.pk})
        self.assertRedirects(
            response,
            reverse('case-attachment', args=[self.case_1.pk]))

    @patch('tcms.core.files.settings.MAX_UPLOAD_SIZE', new=10)
    def test_refuse_if_file_is_too_big(self):
        self.client.login(username=self.user.username, password=self.password)

        with open(self.upload_filename, 'r') as upload_file:
            response = self.client.post(self.upload_file_url,
                                        {'to_plan_id': self.plan.pk,
                                         'upload_file': upload_file})

        self.assertContains(response, 'You upload entity is too large')

    def test_upload_file_to_plan(self):
        self.client.login(username=self.user.username, password=self.password)

        with patch('tcms.core.files.settings.FILE_UPLOAD_DIR',
                   new=self.file_upload_dir):
            with open(self.upload_filename, 'r') as upload_file:
                response = self.client.post(self.upload_file_url,
                                            {'to_plan_id': self.plan.pk,
                                             'upload_file': upload_file})

        self.assertRedirects(
            response,
            reverse('plan-attachment', args=[self.plan.pk]))

        attachments = list(TestAttachment.objects.filter(
            file_name=os.path.basename(self.upload_filename)))
        self.assertTrue(attachments)

        attachment = attachments[0]
        self.assertEqual(self.user.pk, attachment.submitter.pk)

        plan_attachment_rel_exists = TestPlanAttachment.objects.filter(
            plan=self.plan, attachment=attachment).exists()
        self.assertTrue(plan_attachment_rel_exists)

    def test_upload_file_to_case(self):
        self.client.login(username=self.user.username, password=self.password)

        with patch('tcms.core.files.settings.FILE_UPLOAD_DIR',
                   new=self.file_upload_dir):
            with open(self.upload_filename, 'r') as upload_file:
                response = self.client.post(self.upload_file_url,
                                            {'to_case_id': self.case_1.pk,
                                             'upload_file': upload_file})

        self.assertRedirects(
            response,
            reverse('case-attachment', args=[self.case_1.pk]))

        attachments = list(TestAttachment.objects.filter(
            file_name=os.path.basename(self.upload_filename)))
        self.assertTrue(attachments)

        attachment = attachments[0]
        self.assertEqual(self.user.pk, attachment.submitter.pk)

        case_attachment_rel_exists = TestCaseAttachment.objects.filter(
            case=self.case_1, attachment=attachment).exists()
        self.assertTrue(case_attachment_rel_exists)


class TestAbleToDeleteFile(BasePlanCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()

        cls.superuser = f.UserFactory(username='admin')
        cls.superuser.is_superuser = True
        cls.superuser.set_password('admin')
        cls.superuser.save()

        cls.anyone_else = f.UserFactory()

        cls.attachment = f.TestAttachmentFactory()

    def setUp(self):
        super().setUp()

        self.request = RequestFactory()

    def test_superuser_can(self):
        request = self.request.get(
            reverse('delete-file', args=[self.attachment.pk]))
        request.user = self.superuser
        self.assertTrue(able_to_delete_attachment(request, self.attachment.pk))

    def test_attachment_submitter_can(self):
        request = self.request.get(
            reverse('delete-file', args=[self.attachment.pk]))
        request.user = self.attachment.submitter
        self.assertTrue(able_to_delete_attachment(request, self.attachment.pk))

    def test_plan_author_can(self):
        request = self.request.get(
            reverse('delete-file', args=[self.attachment.pk]),
            data={'from_plan': self.plan.pk})
        request.user = self.plan.author
        self.assertTrue(able_to_delete_attachment(request, self.attachment.pk))

    def test_plan_owner_can(self):
        request = self.request.get(
            reverse('delete-file', args=[self.attachment.pk]),
            data={'from_plan': self.plan.pk})
        request.user = self.plan.owner
        self.assertTrue(able_to_delete_attachment(request, self.attachment.pk))

    def test_case_owner_can(self):
        request = self.request.get(
            reverse('delete-file', args=[self.attachment.pk]),
            data={'from_case': self.case_1.pk})
        request.user = self.case_1.author
        self.assertTrue(able_to_delete_attachment(request, self.attachment.pk))

    def test_cannot_delete_by_others(self):
        request = self.request.get(
            reverse('delete-file', args=[self.attachment.pk]),
            data={'from_case': self.case_1.pk})
        request.user = self.anyone_else
        self.assertFalse(able_to_delete_attachment(request, self.attachment.pk))


class TestDeleteFileAuthorization(BasePlanCase):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()

        cls.superuser = f.UserFactory(username='admin')
        cls.superuser.set_password('admin')
        cls.superuser.save()

        cls.anyone_else = f.UserFactory()
        cls.anyone_else_pwd = 'anyone'
        cls.anyone_else.set_password(cls.anyone_else_pwd)
        cls.anyone_else.save()

        cls.plan_attachment = f.TestAttachmentFactory()
        cls.plan_attachment_rel = f.TestPlanAttachmentFactory(
            plan=cls.plan,
            attachment=cls.plan_attachment)
        cls.submitter_pwd = 'secret'
        cls.plan_attachment.submitter.set_password(cls.submitter_pwd)
        cls.plan_attachment.submitter.save()

        cls.case_attachment = f.TestAttachmentFactory()
        cls.case_attachment_rel = f.TestCaseAttachmentFactory(
            case=cls.case_1,
            attachment=cls.case_attachment)
        cls.case_attachment.submitter.set_password(cls.submitter_pwd)
        cls.case_attachment.submitter.save()

    def test_refuse_if_user_cannot_delete_file(self):
        self.client.login(username=self.anyone_else.username,
                          password=self.anyone_else_pwd)

        url = reverse('delete-file', args=[self.plan_attachment.pk])
        response = self.client.get(url, {'from_plan': self.plan.pk})

        self.assertEqual({'rc': 2, 'response': 'auth_failure'}, json.loads(response.content))

    def test_delete_attachment_from_plan(self):
        self.client.login(username=self.plan_attachment.submitter.username,
                          password=self.submitter_pwd)

        url = reverse('delete-file', args=[self.plan_attachment.pk])
        response = self.client.get(url, {'from_plan': self.plan.pk})

        self.assertEqual({'rc': 0, 'response': 'ok'}, json.loads(response.content))
        still_has = self.plan.attachment.filter(pk=self.plan_attachment.pk).exists()
        self.assertFalse(still_has)
        # TODO: skip because delete_file does not delete a TestAttachment object from database
        # self.assertFalse(TestAttachment.objects.filter(pk=self.plan_attachment.pk).exists())

    def test_delete_attachment_from_case(self):
        self.client.login(username=self.case_attachment.submitter.username,
                          password=self.submitter_pwd)

        url = reverse('delete-file', args=[self.case_attachment.pk])
        response = self.client.get(url, {'from_case': self.case_1.pk})

        self.assertEqual({'rc': 0, 'response': 'ok'}, json.loads(response.content))
        still_has = self.case_1.attachment.filter(pk=self.case_attachment.pk).exists()
        self.assertFalse(still_has)
        # TODO: skip because delete_file does not delete a TestAttachment object from database
        # self.assertFalse(TestAttachment.objects.filter(pk=self.case_attachment.pk).exists())


class TestCheckFile(BasePlanCase):
    """Test view method check_file to download an attachment file"""

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()

        cls.upload_dir = tempfile.mkdtemp()

        cls.text_file_content = 'hello Nitrate'
        with open(os.path.join(cls.upload_dir, 'a-1.txt'), 'w', encoding='utf-8') as f:
            f.write(cls.text_file_content)

        cls.binary_file_content = b'\x00\x01\x11\x10'
        with open(os.path.join(cls.upload_dir, 'b.bin'), 'wb') as f:
            f.write(cls.binary_file_content)

        cls.logo_png_content = b'\x00\x01\x10'
        with open(os.path.join(cls.upload_dir, 'logo.png'), 'wb') as f:
            f.write(cls.logo_png_content)

        cls.text_file = TestAttachment.objects.create(
            submitter_id=cls.tester.id,
            description='description',
            file_name='a.txt',
            stored_name='a-1.txt',
            create_date=datetime.now(),
            mime_type='text/plain'
        )
        cls.binary_file = TestAttachment.objects.create(
            submitter_id=cls.tester.id,
            description='binary file',
            file_name='b.txt',
            stored_name='b.bin',
            create_date=datetime.now(),
            mime_type='application/x-binary'
        )
        cls.logo_png = TestAttachment.objects.create(
            submitter_id=cls.tester.id,
            description='binary file',
            file_name='logo.png',
            # stored_name is not set, use file_name to find out attachment instead.
            stored_name=None,
            create_date=datetime.now(),
            mime_type='image/png'
        )
        cls.file_deleted = TestAttachment.objects.create(
            submitter_id=cls.tester.id,
            description='case plan',
            file_name='case-plan.txt',
            stored_name=None,
            create_date=datetime.now(),
            mime_type='text/plain'
        )

    def test_file_id_does_not_exist(self):
        # Calculate a non-existing attachment id. If there is no attachment in
        # database, 1 is expected.
        file_id = (TestAttachment.objects.aggregate(max_id=Max('pk'))['max_id'] or 0) + 1
        resp = self.client.get(reverse('check-file', args=[file_id]))
        self.assert404(resp)

    def test_download_text_file(self):
        with patch.object(settings, 'FILE_UPLOAD_DIR', self.upload_dir):
            resp = self.client.get(reverse('check-file', args=[self.text_file.pk]))
        self.assertEqual('text/plain', resp['Content-Type'])
        self.assertEqual('attachment; filename="a.txt"', resp['Content-Disposition'])
        self.assertEqual(self.text_file_content, resp.content.decode('utf-8'))

    def test_download_binary_file(self):
        with patch.object(settings, 'FILE_UPLOAD_DIR', self.upload_dir):
            resp = self.client.get(reverse('check-file', args=[self.binary_file.pk]))
        self.assertEqual('application/x-binary', resp['Content-Type'])
        self.assertEqual('attachment; filename="b.txt"', resp['Content-Disposition'])
        self.assertEqual(self.binary_file_content, resp.content)

    def test_use_original_filename_to_find_out_attachment(self):
        with patch.object(settings, 'FILE_UPLOAD_DIR', self.upload_dir):
            resp = self.client.get(reverse('check-file', args=[self.logo_png.pk]))
        self.assertEqual('image/png', resp['Content-Type'])
        self.assertEqual('attachment; filename="logo.png"', resp['Content-Disposition'])
        self.assertEqual(self.logo_png_content, resp.content)

    def test_attachment_file_is_deleted_yet(self):
        with patch.object(settings, 'FILE_UPLOAD_DIR', self.upload_dir):
            resp = self.client.get(reverse('check-file', args=[self.file_deleted.pk]))
        self.assert404(resp)



# -*- coding: utf-8 -*-

import unittest

from django import test

from tcms.testcases.models import TestCase
from tcms.testcases.models import TestCasePlan
from tcms.testplans.models import TCMSEnvPlanMap
from tcms.testplans.models import TestPlan
from tcms.xmlrpc.api import testplan as XmlrpcTestPlan
from tcms.xmlrpc.api.testplan import import_case_via_XML
from tcms.xmlrpc.serializer import datetime_to_str
from tests import factories as f
from tests.testplans.test_importer import xml_file_without_error
from tests.xmlrpc.utils import make_http_request
from tests.xmlrpc.utils import XmlrpcAPIBaseTest

__all__ = (
    'TestAddComponent',
    'TestAddTag',
    'TestComponentMethods',
    'TestFilter',
    'TestGetAllCasesTags',
    'TestGetProduct',
    'TestGetTestCases',
    'TestGetTestRuns',
    'TestGetText',
    'TestImportCaseViaXML',
    'TestPlanTypeMethods',
    'TestRemoveTag',
    'TestUpdate',
)


class TestFilter(XmlrpcAPIBaseTest):

    @classmethod
    def setUpTestData(cls):
        cls.product = f.ProductFactory()
        cls.version = f.VersionFactory(product=cls.product)
        cls.tester = f.UserFactory()
        cls.plan_type = f.TestPlanTypeFactory(name='manual smoking')
        cls.plan_1 = f.TestPlanFactory(
            product_version=cls.version,
            product=cls.product,
            author=cls.tester,
            type=cls.plan_type)
        cls.plan_2 = f.TestPlanFactory(
            product_version=cls.version,
            product=cls.product,
            author=cls.tester,
            type=cls.plan_type)
        cls.case_1 = f.TestCaseFactory(
            author=cls.tester,
            default_tester=None,
            reviewer=cls.tester,
            plan=[cls.plan_1])
        cls.case_2 = f.TestCaseFactory(
            author=cls.tester,
            default_tester=None,
            reviewer=cls.tester,
            plan=[cls.plan_1])

    def test_filter_plans(self):
        plans = XmlrpcTestPlan.filter(None, {'pk__in': [self.plan_1.pk, self.plan_2.pk]})
        plan = plans[0]
        self.assertEqual(self.plan_1.name, plan['name'])
        self.assertEqual(self.plan_1.product_version.pk, plan['product_version_id'])
        self.assertEqual(self.plan_1.author.pk, plan['author_id'])

        self.assertEqual(2, len(plan['case']))
        self.assertEqual([self.case_1.pk, self.case_2.pk], plan['case'])
        self.assertEqual(0, len(plans[1]['case']))

    def test_filter_out_all_plans(self):
        plans_total = TestPlan.objects.all().count()
        self.assertEqual(plans_total, len(XmlrpcTestPlan.filter(None)))
        self.assertEqual(plans_total, len(XmlrpcTestPlan.filter(None, {})))


class TestAddTag(XmlrpcAPIBaseTest):

    @classmethod
    def setUpTestData(cls):
        cls.user = f.UserFactory()
        cls.http_req = make_http_request(user=cls.user,
                                         user_perm='testplans.add_testplantag')

        cls.product = f.ProductFactory()
        cls.plans = [
            f.TestPlanFactory(
                author=cls.user, owner=cls.user, product=cls.product),
            f.TestPlanFactory(
                author=cls.user, owner=cls.user, product=cls.product),
        ]

        cls.tag1 = f.TestTagFactory(name='xmlrpc_test_tag_1')
        cls.tag2 = f.TestTagFactory(name='xmlrpc_test_tag_2')
        cls.tag_name = 'xmlrpc_tag_name_1'

    def test_single_id(self):
        """Test with singal plan id and tag id"""
        self.assertXmlrpcFaultInternalServerError(
            XmlrpcTestPlan.add_tag,
            self.http_req, self.plans[0].pk, self.tag1.pk)

        XmlrpcTestPlan.add_tag(self.http_req, self.plans[0].pk, self.tag1.name)
        tag_exists = TestPlan.objects.filter(pk=self.plans[0].pk, tag__pk=self.tag1.pk).exists()
        self.assertTrue(tag_exists)

    def test_array_argument(self):
        XmlrpcTestPlan.add_tag(self.http_req, self.plans[0].pk, [self.tag2.name, self.tag_name])
        tag_exists = TestPlan.objects.filter(pk=self.plans[0].pk,
                                             tag__name__in=[self.tag2.name, self.tag_name])
        self.assertTrue(tag_exists.exists())

        plans_ids = [plan.pk for plan in self.plans]
        tags_names = [self.tag_name, 'xmlrpc_tag_name_2']
        XmlrpcTestPlan.add_tag(self.http_req, plans_ids, tags_names)
        for plan in self.plans:
            tag_exists = plan.tag.filter(name__in=tags_names).exists()
            self.assertTrue(tag_exists)


class TestAddComponent(XmlrpcAPIBaseTest):

    @classmethod
    def setUpTestData(cls):
        cls.user = f.UserFactory()
        cls.http_req = make_http_request(user=cls.user,
                                         user_perm='testplans.add_testplancomponent')

        cls.product = f.ProductFactory()
        cls.plans = [
            f.TestPlanFactory(
                author=cls.user, owner=cls.user, product=cls.product),
            f.TestPlanFactory(
                author=cls.user, owner=cls.user, product=cls.product),
        ]
        cls.component1 = f.ComponentFactory(
            name='xmlrpc test component 1',
            description='xmlrpc test description',
            product=cls.product,
            initial_owner=None,
            initial_qa_contact=None)
        cls.component2 = f.ComponentFactory(
            name='xmlrpc test component 2',
            description='xmlrpc test description',
            product=cls.product,
            initial_owner=None,
            initial_qa_contact=None)

    def test_single_id(self):
        XmlrpcTestPlan.add_component(self.http_req, self.plans[0].pk, self.component1.pk)
        component_exists = TestPlan.objects.filter(
            pk=self.plans[0].pk, component__pk=self.component1.pk).exists()
        self.assertTrue(component_exists)

    def test_ids_in_array(self):
        self.assertXmlrpcFaultBadRequest(
            XmlrpcTestPlan.add_component, self.http_req, [1, 2])

        plans_ids = [plan.pk for plan in self.plans]
        components_ids = [self.component1.pk, self.component2.pk]
        XmlrpcTestPlan.add_component(self.http_req, plans_ids, components_ids)
        for plan in TestPlan.objects.filter(pk__in=plans_ids):
            components_ids = [item.pk for item in plan.component.iterator()]
            self.assertTrue(self.component1.pk in components_ids)
            self.assertTrue(self.component2.pk in components_ids)


class TestPlanTypeMethods(XmlrpcAPIBaseTest):

    @classmethod
    def setUpTestData(cls):
        cls.http_req = make_http_request()
        cls.plan_type = f.TestPlanTypeFactory(
            name='xmlrpc plan type', description='')

    def test_check_plan_type(self):
        self.assertXmlrpcFaultBadRequest(
            XmlrpcTestPlan.check_plan_type, self.http_req)

        result = XmlrpcTestPlan.check_plan_type(self.http_req, self.plan_type.name)
        self.assertEqual(self.plan_type.name, result['name'])
        self.assertEqual(self.plan_type.description, result['description'])
        self.assertEqual(self.plan_type.pk, result['id'])

    def test_get_plan_type(self):
        result = XmlrpcTestPlan.get_plan_type(self.http_req, self.plan_type.pk)
        self.assertEqual(self.plan_type.name, result['name'])
        self.assertEqual(self.plan_type.description, result['description'])
        self.assertEqual(self.plan_type.pk, result['id'])

        self.assertXmlrpcFaultNotFound(
            XmlrpcTestPlan.get_plan_type, self.http_req, 0)
        self.assertXmlrpcFaultNotFound(
            XmlrpcTestPlan.get_plan_type, self.http_req, -2)


class TestGetProduct(XmlrpcAPIBaseTest):

    @classmethod
    def setUpTestData(cls):
        cls.http_req = make_http_request()
        cls.user = f.UserFactory()
        cls.product = f.ProductFactory()
        cls.plan = f.TestPlanFactory(
            author=cls.user, owner=cls.user, product=cls.product)

    def _verify_serialize_result(self, result):
        self.assertEqual(self.plan.product.name, result['name'])
        self.assertEqual(self.plan.product.description, result['description'])
        self.assertEqual(self.plan.product.disallow_new, result['disallow_new'])
        self.assertEqual(self.plan.product.vote_super_user, result['vote_super_user'])
        self.assertEqual(self.plan.product.max_vote_super_bug, result['max_vote_super_bug'])
        self.assertEqual(self.plan.product.votes_to_confirm, result['votes_to_confirm'])
        self.assertEqual(self.plan.product.default_milestone, result['default_milestone'])
        self.assertEqual(self.plan.product.classification.pk, result['classification_id'])
        self.assertEqual(self.plan.product.classification.name, result['classification'])

    def test_get_product(self):
        # self.assertXmlrpcFaultNotFound(
        #     XmlrpcTestPlan.get_product, self.http_req, 0)

        result = XmlrpcTestPlan.get_product(self.http_req, str(self.plan.pk))
        self._verify_serialize_result(result)

        result = XmlrpcTestPlan.get_product(self.http_req, self.plan.pk)
        self._verify_serialize_result(result)

        self.assertXmlrpcFaultBadRequest(
            XmlrpcTestPlan.get_product, self.http_req, 'plan_id')


@unittest.skip('TODO: test case is not implemented yet.')
class TestComponentMethods(test.TestCase):
    """TODO: """


@unittest.skip('TODO: test case is not implemented yet.')
class TestGetAllCasesTags(test.TestCase):
    """TODO: """


class TestGetTestCases(XmlrpcAPIBaseTest):
    """Test testplan.get_test_cases method"""

    @classmethod
    def setUpTestData(cls):
        cls.http_req = make_http_request()

        cls.tester = f.UserFactory(username='tester')
        cls.reviewer = f.UserFactory(username='reviewer')
        cls.product = f.ProductFactory()
        cls.plan = f.TestPlanFactory(
            author=cls.tester, owner=cls.tester, product=cls.product)
        cls.cases = [
            f.TestCaseFactory(author=cls.tester,
                              default_tester=None,
                              reviewer=cls.reviewer,
                              plan=[cls.plan]),
            f.TestCaseFactory(author=cls.tester,
                              default_tester=None,
                              reviewer=cls.reviewer,
                              plan=[cls.plan]),
            f.TestCaseFactory(author=cls.tester,
                              default_tester=None,
                              reviewer=cls.reviewer,
                              plan=[cls.plan]),
        ]
        cls.another_plan = f.TestPlanFactory(
            author=cls.tester, owner=cls.tester, product=cls.product)

    def test_get_test_cases(self):
        serialized_cases = XmlrpcTestPlan.get_test_cases(self.http_req, self.plan.pk)
        for case in serialized_cases:
            expected_case = TestCase.objects.filter(plan=self.plan.pk).get(pk=case['case_id'])

            self.assertEqual(expected_case.summary, case['summary'])
            self.assertEqual(expected_case.priority_id, case['priority_id'])
            self.assertEqual(expected_case.author_id, case['author_id'])

            plan_case_rel = TestCasePlan.objects.get(plan=self.plan, case=case['case_id'])
            self.assertEqual(plan_case_rel.sortkey, case['sortkey'])

    @unittest.skip('TODO: fix get_test_cases to make this test pass.')
    def test_different_argument_type(self):
        self.assertXmlrpcFaultBadRequest(
            XmlrpcTestPlan.get_test_cases, self.http_req, str(self.plan.pk))

    def test_404_when_plan_nonexistent(self):
        self.assertXmlrpcFaultNotFound(
            XmlrpcTestPlan.get_test_cases, self.http_req, 0)

        plan_id = TestPlan.objects.order_by('-pk')[:1][0].pk + 1
        self.assertXmlrpcFaultNotFound(
            XmlrpcTestPlan.get_test_cases, self.http_req, plan_id)

    def test_plan_has_no_cases(self):
        result = XmlrpcTestPlan.get_test_cases(self.http_req, self.another_plan.pk)
        self.assertEqual([], result)


@unittest.skip('TODO: test case is not implemented yet.')
class TestGetTestRuns(test.TestCase):
    """TODO: """


@unittest.skip('TODO: test case is not implemented yet.')
class TestGetText(test.TestCase):
    """TODO: """


@unittest.skip('TODO: test case is not implemented yet.')
class TestRemoveTag(test.TestCase):
    """TODO: """


class TestUpdate(test.TestCase):
    """ Tests the XMLRPM testplan.update method """

    @classmethod
    def setUpTestData(cls):
        cls.user = f.UserFactory()
        cls.http_req = make_http_request(user=cls.user,
                                         user_perm='testplans.change_testplan')

        cls.env_group_1 = f.TCMSEnvGroupFactory()
        cls.env_group_2 = f.TCMSEnvGroupFactory()
        cls.product = f.ProductFactory()
        cls.version = f.VersionFactory(product=cls.product)
        cls.tester = f.UserFactory()
        cls.plan_type = f.TestPlanTypeFactory(name='manual smoking')
        cls.plan_1 = f.TestPlanFactory(
            product_version=cls.version,
            product=cls.product,
            author=cls.tester,
            type=cls.plan_type,
            env_group=(cls.env_group_1,))
        cls.plan_2 = f.TestPlanFactory(
            product_version=cls.version,
            product=cls.product,
            author=cls.tester,
            type=cls.plan_type,
            env_group=(cls.env_group_1,))

    def test_update_env_group(self):
        # plan_1 and plan_2 point to self.env_group_1
        # and there are only 2 objects in the many-to-many table
        # so we issue XMLRPC request to modify the env_group of self.plan_2
        plans = XmlrpcTestPlan.update(self.http_req, self.plan_2.pk, {'env_group': self.env_group_2.pk})
        plan = plans[0]

        # now verify that the returned TP (plan_2) has been updated to env_group_2
        self.assertEqual(self.plan_2.pk, plan['plan_id'])
        self.assertEqual(1, len(plan['env_group']))
        self.assertEqual(self.env_group_2.pk, plan['env_group'][0])

        # and that plan_1 has not changed at all
        self.assertEqual(1, self.plan_1.env_group.count())
        self.assertEqual(self.env_group_1.pk, self.plan_1.env_group.all()[0].pk)

        # and there are still only 2 objects in the many-to-many table
        # iow no dangling objects left
        self.assertEqual(2, TCMSEnvPlanMap.objects.filter(plan__in=[self.plan_1, self.plan_2]).count())


class TestImportCaseViaXML(XmlrpcAPIBaseTest):
    """Test import_case_via_XML"""

    @classmethod
    def setUpTestData(cls):
        cls.author = f.UserFactory(username='user', email='user@example.com')
        cls.plan = f.TestPlanFactory(author=cls.author, owner=cls.author)
        cls.http_req = make_http_request(user=cls.author)

    def test_succeed_to_import_cases(self):
        result = import_case_via_XML(self.http_req,
                                     self.plan.pk,
                                     xml_file_without_error)

        self.assertEqual(2, TestCase.objects.count())
        self.assertTrue(TestCase.objects.filter(summary='case 2').exists())
        self.assertEqual('Success update 2 cases', result)


class TestGet(test.TestCase):
    """Test TestPlan.get"""

    @classmethod
    def setUpTestData(cls):
        cls.author = f.UserFactory(username='user', email='user@example.com')
        cls.http_req = make_http_request(user=cls.author)

        cls.product = f.ProductFactory()
        cls.version = f.VersionFactory(product=cls.product)
        cls.type = f.TestPlanTypeFactory(name='temp')
        cls.tag_fedora = f.TestTagFactory(name='fedora')
        cls.tag_centos = f.TestTagFactory(name='centos')

        cls.plan = f.TestPlanFactory(
            is_active=True,
            extra_link=None,
            product=cls.product,
            product_version=cls.version,
            owner=cls.author,
            author=cls.author,
            parent=None,
            type=cls.type,
            tag=[cls.tag_fedora, cls.tag_centos]
        )

    def test_get(self):
        self.maxDiff = None
        expected_plan = {
            'plan_id': self.plan.pk,
            'name': self.plan.name,
            'create_date': datetime_to_str(self.plan.create_date),
            'is_active': True,
            'extra_link': None,
            'product_version_id': self.version.pk,
            'product_version': self.version.value,
            'default_product_version': self.version.value,
            'owner_id': self.author.pk,
            'owner': self.author.username,
            'author_id': self.author.pk,
            'author': self.author.username,
            'product_id': self.product.pk,
            'product': self.product.name,
            'type_id': self.type.pk,
            'type': self.type.name,
            'parent_id': None,
            'parent': None,
            'attachment': [],
            'component': [],
            'env_group': [],
            'tag': ['centos', 'fedora']
        }

        plan = XmlrpcTestPlan.get(self.http_req, self.plan.pk)
        plan['tag'].sort()
        self.assertEqual(expected_plan, plan)



# -*- coding: utf-8 -*-

from django.contrib.auth.models import User
from django.test import TestCase
from django.urls import reverse

from tcms.profiles.forms import BookmarkForm


class TestOpenBookmarks(TestCase):
    """Test for opening bookmarks"""

    @classmethod
    def setUpClass(cls):
        super().setUpClass()

        cls.tester = User.objects.create_user(username='bookmark_tester',
                                              email='bookmark_tester@example.com',
                                              password='password')

        bookmark_form = BookmarkForm({
            'name': 'plan page',
            'url': 'http://localhost/plan/1/',
            'user': cls.tester.pk,
            'a': 'add',
        })
        bookmark_form.is_valid()
        cls.bookmark_1 = bookmark_form.save()

        bookmark_form = BookmarkForm({
            'name': 'case page',
            'url': 'http://localhost/case/1/',
            'user': cls.tester.pk,
            'a': 'add',
        })
        bookmark_form.is_valid()
        cls.bookmark_2 = bookmark_form.save()

        bookmark_form = BookmarkForm({
            'name': 'run page',
            'url': 'http://localhost/run/1/',
            'user': cls.tester.pk,
            'a': 'add',
        })
        bookmark_form.is_valid()
        cls.bookmark_3 = bookmark_form.save()

    def test_open_bookmark_page(self):
        self.client.login(username=self.tester.username, password='password')

        url = reverse('user-bookmark',
                      kwargs={'username': self.tester.username})
        response = self.client.get(url)

        checkbox_fmt = '<input name="pk" value="{}" type="checkbox" />'

        for bookmark in (self.bookmark_1, self.bookmark_2, self.bookmark_3):
            checkbox = checkbox_fmt.format(bookmark.pk)
            self.assertContains(response, checkbox, html=True)

        self.assertContains(
            response,
            '<input value="Delete" class="sprites node_delete" type="submit">',
            html=True)



# -*- coding: utf-8 -*-

from django.conf.urls import url
from .. import views

urlpatterns = [
    url(r'^new/$', views.new, name='cases-new'),
    url(r'^$', views.all, name='cases-all'),
    url(r'^search/$', views.search, name='cases-search'),
    url(r'^load-more/$', views.load_more_cases, name='cases-load-more'),
    url(r'^ajax/$', views.ajax_search, name='cases-ajax-search'),

    url(r'^automated/$', views.ChangeCaseAutomatedPropertyView.as_view(),
        name='cases-automated'),

    url(r'^tag/$', views.tag, name='cases-tag'),
    url(r'^category/$', views.category, name='cases-category'),
    url(r'^clone/$', views.clone, name='cases-clone'),
    url(r'^printable/$', views.printable, name='cases-printable'),
    url(r'^export/$', views.export, name='cases-export'),

    url(r'^add-component/$', views.AddComponentView.as_view(),
        name='cases-add-component'),
    url(r'^remove-component/$', views.RemoveComponentView.as_view(),
        name='cases-remove-component'),
    url(r'^get-component-form/$', views.GetComponentFormView.as_view(),
        name='cases-get-component-form'),
]



# -*- coding: utf-8 -*-

import logging

from django.conf import settings
from django.contrib.auth.mixins import PermissionRequiredMixin
from django.http import JsonResponse
from django.shortcuts import render
from django.views import generic
from django.views.decorators.http import require_POST

import django_comments.signals
from django_comments.views.moderation import perform_delete

import tcms.comments
from tcms.core import utils

log = logging.getLogger(__name__)


class InvalidCommentPostRequest(Exception):
    """Raised if comment is not valid"""

    def __init__(self, target, form):
        self.target = target
        self.form = form


def get_comment_target_object(content_type: str, object_pk: int):
    """Get corresponding model object

    :param content_type: a string representing the model's content type in
        format ``app_name.model_name``, for example, ``testcases.testcase``.
    :type content_type: str
    :param object_pk: the pk of the object.
    :type object_pk: int
    :return: the model object according to the content type.
    """
    return utils.get_model(content_type)._default_manager.get(pk=object_pk)


def post_comment(comment_data, request_user, remote_addr=None):
    """Save a comment

    This function is a customized version of
    ``django_comments.views.comments.post_comment`` by cutting off part of the
    code that are not necessary for Nitrate, for example, this method does not
    send signal before and after saving comment. A significant change between
    them is this function is not a view function and can be reused to add a
    comment to a given model object.
    """

    if request_user.is_authenticated:
        # Nitrate does not allow user to enter name or email for adding a
        # comment, so when the request is authenticated, use the logged-in
        # user's name and email.
        comment_data["name"] = request_user.get_full_name() or request_user.username
        comment_data["email"] = request_user.email

    ctype = comment_data.get('content_type')
    object_pk = comment_data.get('object_pk')

    target = get_comment_target_object(ctype, int(object_pk))

    form = tcms.comments.get_form()(target, data=comment_data)
    if not form.is_valid():
        log.error('Failed to add comment to %s %s: %r',
                  ctype, object_pk, comment_data)
        log.error('Error messages: %r', form.errors)
        raise InvalidCommentPostRequest(target, form)

    # Otherwise create the comment
    comment = form.get_comment_object()
    comment.ip_address = remote_addr
    if request_user.is_authenticated:
        comment.user = request_user
    comment.save()

    return target, comment


@require_POST
def post(request, template_name='comments/comments.html'):
    """Post a comment"""
    data = request.POST.copy()
    try:
        target, _ = post_comment(
            data, request.user, request.META.get('REMOTE_ADDR'))
    except InvalidCommentPostRequest as e:
        target = e.target
    return render(request, template_name, context={'object': target})


class DeleteCommentView(PermissionRequiredMixin, generic.View):
    """Delete comment from given objects"""

    permission_required = 'django_comments.can_moderate'

    def post(self, request):
        comments = django_comments.get_model().objects.filter(
            pk__in=request.POST.getlist('comment_id'),
            site__pk=settings.SITE_ID,
            is_removed=False,
            user_id=request.user.id
        )

        if not comments:
            return JsonResponse({'rc': 1, 'response': 'Object does not exist.'})

        # Flag the comment as deleted instead of actually deleting it.
        for comment in comments:
            perform_delete(request, comment)

        return JsonResponse({'rc': 0, 'response': 'ok'})



# -*- coding: utf-8 -*-

import logging

import tcms.comments
from django.core.exceptions import ObjectDoesNotExist
from tcms.comments.views import (
    get_comment_target_object, post_comment, InvalidCommentPostRequest
)

log = logging.getLogger(__name__)


def add_comment(
        request_user, content_type, object_pks, comment, remote_addr=None):
    """Add comment to given target objects

    This function is useful particularly for not add a comment from WebUI by
    clicking a button, and allow to add a comment to more than one model
    objects at once.

    :param request_user: the user object who requests to add this comment.
    :type request_user: :class:`django.contrib.auth.models.User`
    :param str content_type: the content type of the model in format
        ``app_label.model_name``, e.g. ``testcases.testcase``.
    :param object_pks: list of object IDs in order to allow adding a comment to
        more than on models.
    :type object_pks: list[int]
    :param str comment: content of the comment to be added.
    :param remote_addr: remote IP address of the user.
    :type remote_addr: str or None
    :return: list of comments just added.
    :rtype: list[:class:`django_comments.models.Comment`]
    """
    new_comments = []
    comment_form = tcms.comments.get_form()
    for object_pk in object_pks:
        try:
            target = get_comment_target_object(content_type, object_pk)
        except ObjectDoesNotExist:
            log.error('%s object with id %s does not exist in database. '
                      'Ignore it and continue to add comment to next one.',
                      content_type, object_pk)
            continue
        initial_form = comment_form(target)
        comment_data = initial_form.initial.copy()
        comment_data['comment'] = comment
        try:
            _, new_comment = post_comment(
                comment_data, request_user, remote_addr)
        except InvalidCommentPostRequest:
            log.error(
                'Failed to add comment, continue to add comment to next one.')
        else:
            new_comments.append(new_comment)
    return new_comments



# -*- coding: utf-8 -*-

from django.conf.urls import url
from django.views.generic.base import RedirectView

from . import views

urlpatterns = [
    url(r'^$', RedirectView.as_view(url='overall/', permanent=True)),
    url(r'^overall/$', views.overall, name='report-overall'),
    url(r'^product/(?P<product_id>\d+)/overview/$', views.overview,
        name='report-overview'),
    url(r'^product/(?P<product_id>\d+)/build/$',
        views.ProductBuildReport.as_view(),
        name='report-overall-product-build'),
    url(r'^product/(?P<product_id>\d+)/version/$',
        views.ProductVersionReport.as_view(),
        name='report-overall-product-version'),
    url(r'^product/(?P<product_id>\d+)/component/$',
        views.ProductComponentReport.as_view(),
        name='report-overall-product-component'),
    url(r'^custom/$', views.CustomReport.as_view(), name='report-custom'),
    url(r'^custom/details/$', views.CustomDetailReport.as_view(),
        name='report-custom-details'),

    url(r'^testing/$', views.TestingReport.as_view(), name='testing-report'),
    url(r'^testing/case-runs/$',
        views.TestingReportCaseRuns.as_view(),
        name='testing-report-case-runs'),
]



# -*- coding: utf-8 -*-

from django.conf.urls import url
from . import ajax, files, views

urlpatterns = [
    # Site entry
    url(r'^$', views.index, name='nitrate-index'),
    url(r'^search/$', views.search, name='nitrate-search'),

    # Ajax call responder
    url(r'^ajax/update/$', ajax.update, name='ajax-update'),

    # TODO: merge this into next mapping
    url(r'^ajax/update/case-status/$', ajax.update_cases_case_status),
    url(r'^ajax/update/case-run-status$', ajax.update_case_run_status,
        name='ajax-update-caserun-status'),
    url(r'^ajax/update/cases-priority/$', ajax.update_cases_priority),
    url(r'^ajax/update/cases-default-tester/$', ajax.update_cases_default_tester,
        name='ajax-update-cases-default-tester'),
    url(r'^ajax/update/cases-reviewer/$', ajax.update_cases_reviewer),
    url(r'^ajax/update/cases-sortkey/$', ajax.update_cases_sortkey),
    url(r'^ajax/form/$', ajax.form, name='ajax-form'),
    url(r'^ajax/get-prod-relate-obj/$', ajax.get_prod_related_obj_json),
    url(r'^management/getinfo/$', ajax.info, name='ajax-getinfo'),
    url(r'^management/tags/$', ajax.tag),

    # Attached file zone
    url(r'^management/uploadfile/$', files.UploadFileView.as_view(), name='upload-file'),
    url(r'^management/checkfile/(?P<file_id>\d+)/$', files.check_file,
        name='check-file'),
    url(r'^management/deletefile/(?P<file_id>\d+)/$', files.delete_file,
        name='delete-file'),
]



# -*- coding: utf-8 -*-

from django.conf.urls import url
from tcms.linkreference import views

urlpatterns = [
    url(r'^add/$', views.AddLinkToTargetView.as_view(),
        name='add-link-reference'),

    url(r'^get/$', views.get, name='get-link-references'),

    url(r'^remove/(?P<link_id>\d+)/$', views.RemoveLinkReferenceView.as_view(),
        name='remove-link-reference'),
]



# -*- coding: utf-8 -*-
# Download from http://www.djangosnippets.org/snippets/721/
# All right reserved by  the orignal author(s).
from django.template import (
    Library, TemplateSyntaxError, Node, NodeList, Variable,
    VariableDoesNotExist
)

register = Library()


class IfInNode(Node):
    def __init__(self, var1, var2, nodelist_true, nodelist_false, negate):
        self.var1, self.var2 = Variable(var1), Variable(var2)
        self.nodelist_true, self.nodelist_false = nodelist_true, nodelist_false
        self.negate = negate

    def __repr__(self):
        return "<IfInNode>"

    def render(self, context):
        try:
            val1 = self.var1.resolve(context)
        except VariableDoesNotExist:
            val1 = None
        try:
            val2 = self.var2.resolve(context)
        except VariableDoesNotExist:
            val2 = None
        try:
            if ((self.negate and val1 not in val2) or
                    (not self.negate and val1 in val2)):
                return self.nodelist_true.render(context)
            return self.nodelist_false.render(context)
        except TypeError:
            raise ValueError("Second arg to ifin or ifnotin must be iterable")


def do_ifin(parser, token, negate):
    bits = list(token.split_contents())
    if len(bits) != 3:
        raise TemplateSyntaxError("%r takes two arguments" % bits[0])
    end_tag = 'end' + bits[0]
    nodelist_true = parser.parse(('else', end_tag))
    token = parser.next_token()
    if token.contents == 'else':
        nodelist_false = parser.parse((end_tag,))
        parser.delete_first_token()
    else:
        nodelist_false = NodeList()
    return IfInNode(bits[1], bits[2], nodelist_true, nodelist_false, negate)


def ifin(parser, token):
    return do_ifin(parser, token, False)


register.tag('ifin', ifin)


def ifnotin(parser, token):
    return do_ifin(parser, token, True)


register.tag('ifnotin', ifnotin)



# -*- coding: utf-8 -*-

from json import dumps as json_dumps
from itertools import groupby
from operator import itemgetter

from django.contrib.auth.decorators import permission_required
from django.contrib.contenttypes.models import ContentType
from django.db import IntegrityError
from django.http import Http404
from django.http import HttpResponse
from django.http import JsonResponse
from django.shortcuts import render
from django.views.decorators.http import require_GET

from tcms.logs.models import TCMSLogModel
from tcms.core.utils import QuerySetIterationProxy
from tcms.management.models import TCMSEnvGroup
from tcms.management.models import TCMSEnvGroupPropertyMap
from tcms.management.models import TCMSEnvProperty
from tcms.management.models import TCMSEnvValue

MODULE_NAME = "management"


@require_GET
def environment_groups(request, template_name='environment/groups.html'):
    """
    Environements list
    """

    env_groups = TCMSEnvGroup.objects

    has_perm = request.user.has_perm
    user_action = request.GET.get('action')

    # Add action
    if user_action == 'add':
        if not has_perm('management.add_tcmsenvgroup'):
            return JsonResponse({'rc': 1, 'response': 'Permission denied.'})

        group_name = request.GET.get('name')

        # Get the group name of envrionment from javascript
        if not group_name:
            return JsonResponse(
                {'rc': 1, 'response': 'Environment group name is required.'})

        if env_groups.filter(name=group_name).exists():
            response_msg = (
                f"Environment group name '{group_name}' already exists, "
                f"please select another name."
            )
            return JsonResponse({'rc': 1, 'response': response_msg})
        else:
            env = env_groups.create(name=group_name,
                                    manager_id=request.user.id,
                                    modified_by_id=None)
            env.log_action(who=request.user,
                           new_value='Initial env group %s' % env.name)
            return JsonResponse({'rc': 0, 'response': 'ok', 'id': env.id})

    # Del action
    if user_action == 'del':
        if not request.GET.get('id'):
            raise Http404

        try:
            group_pk = int(request.GET['id'])
        except ValueError:
            return JsonResponse({'rc': 1, 'response': 'id must be an integer.'})

        groups = list(env_groups.filter(pk=group_pk))
        if len(groups) == 0:
            raise Http404

        if request.user.pk != groups[0].manager_id:
            if not has_perm('management.delete_tcmsenvgroup'):
                return JsonResponse({'rc': 1, 'response': 'Permission denied.'})

        groups[0].delete()

        return JsonResponse({'rc': 0, 'response': 'ok'})

    # Modify actions
    if user_action == 'modify':
        if not has_perm('management.change_tcmsenvgroup'):
            return JsonResponse({'rc': 1, 'response': 'Permission denied.'})

        try:
            env = env_groups.get(id=request.GET['id'])
            if request.GET.get('status') in ['0', '1']:
                returned_status = bool(int(request.GET['status']))
                if env.is_active != returned_status:
                    env.is_active = returned_status
                    env.save(update_fields=['is_active'])

                    env.log_action(
                        who=request.user,
                        field='is_active',
                        original_value=env.is_active,
                        new_value=returned_status)
            else:
                return JsonResponse({'rc': 1, 'response': 'Argument illegel.'})
        except TCMSEnvGroup.DoesNotExist as error:
            raise Http404(error)

    # Search actions
    if user_action == 'search':
        if request.GET.get('name'):
            env_groups = env_groups.filter(
                name__icontains=request.GET['name']
            )
        else:
            env_groups = env_groups.all()
    else:
        env_groups = env_groups.all().order_by('is_active')

    # Get properties for each group
    qs = TCMSEnvGroupPropertyMap.objects.filter(group__in=env_groups)
    qs = qs.values('group__pk', 'property__name')
    qs = qs.order_by('group__pk', 'property__name').iterator()
    properties = {
        key: list(value) for key, value in groupby(qs, itemgetter('group__pk'))
    }

    # Get logs for each group
    env_group_ct = ContentType.objects.get_for_model(TCMSEnvGroup)
    qs = TCMSLogModel.objects.filter(content_type=env_group_ct,
                                     object_pk__in=env_groups)
    qs = qs.values('object_pk', 'who__username', 'date', 'field',
                   'original_value', 'new_value')
    qs = qs.order_by('object_pk').iterator()
    # we have to convert object_pk to an integer due to it's a string stored in
    # database.
    logs = {
        int(key): list(value) for key, value in
        groupby(qs, itemgetter('object_pk'))
    }

    env_groups = env_groups.select_related('modified_by', 'manager').iterator()

    env_groups = QuerySetIterationProxy(env_groups,
                                        properties=properties,
                                        another_logs=logs)
    context_data = {
        'environments': env_groups,
        'module': 'env',
    }
    return render(request, template_name, context=context_data)


@require_GET
@permission_required('management.change_tcmsenvgroup')
def environment_group_edit(request, template_name='environment/group_edit.html'):
    """
    Assign properties to environment group
    """

    # Initial the response
    response = ''
    environment_id = request.GET.get('id', None)

    if environment_id is None:
        raise Http404

    try:
        environment = TCMSEnvGroup.objects.get(pk=environment_id)
    except TCMSEnvGroup.DoesNotExist:
        raise Http404

    try:
        de = TCMSEnvGroup.objects.get(name=request.GET.get('name'))
        if environment != de:
            context_data = {
                'environment': environment,
                'properties': TCMSEnvProperty.get_active(),
                'selected_properties': environment.property.all(),
                'message': 'Duplicated name already exists, please change to '
                           'another name.',
            }
            return render(request, template_name, context=context_data)
    except TCMSEnvGroup.DoesNotExist:
        pass

    if request.GET.get('action') == 'modify':   # Actions of modify
        environment_name = request.GET['name']
        if environment.name != environment_name:
            original_value = environment.name
            environment.name = environment_name
            environment.log_action(
                who=request.user,
                field='name',
                original_value=original_value,
                new_value=environment_name)

        returned_env_status = 'enabled' in request.GET
        if environment.is_active != returned_env_status:
            original_value = environment.is_active
            environment.is_active = returned_env_status
            environment.log_action(
                who=request.user,
                field='is_active',
                original_value=original_value,
                new_value=returned_env_status)

        environment.modified_by_id = request.user.id
        environment.save()

        original_property_values = list(
            environment.property.values_list('name', flat=True))
        original_property_values.sort()

        # Remove all of properties of the group.
        TCMSEnvGroupPropertyMap.objects.filter(group__id=environment.id).delete()

        # Readd the property to environemnt group and log the action
        for property_id in request.GET.getlist('selected_property_ids'):
            TCMSEnvGroupPropertyMap.objects.create(group_id=environment.id,
                                                   property_id=property_id)

        new_property_values = list(
            environment.property.values_list('name', flat=True))
        new_property_values.sort()

        environment.log_action(
            who=request.user,
            field='Property values',
            original_value=', '.join(original_property_values),
            new_value=', '.join(new_property_values))

        response = 'Environment group saved successfully.'

    context_data = {
        'environment': environment,
        'properties': TCMSEnvProperty.get_active(),
        'selected_properties': environment.property.all(),
        'message': response,
    }
    return render(request, template_name, context=context_data)


@require_GET
def environment_properties(request, template_name='environment/property.html'):
    """
    Edit environemnt properties and values belong to
    """

    # Initial the ajax response
    ajax_response = {'rc': 0, 'response': 'ok'}
    message = ''

    has_perm = request.user.has_perm
    user_action = request.GET.get('action')

    # Actions of create properties
    if user_action == 'add':
        if not has_perm('management.add_tcmsenvproperty'):
            return JsonResponse({'rc': 1, 'response': 'Permission denied'})

        property_name = request.GET.get('name')

        if not property_name:
            return JsonResponse({'rc': 1, 'response': 'Property name is required'})

        if TCMSEnvProperty.objects.filter(name=property_name).exists():
            resp_msg = (
                f"Environment property named '{property_name}' already exists,"
                f" please select another name."
            )
            return JsonResponse({'rc': 1, 'response': resp_msg})

        new_property = TCMSEnvProperty.objects.create(name=property_name)

        return JsonResponse({
            'rc': 0,
            'response': 'ok',
            'id': new_property.pk,
            'name': new_property.name
        })

    # Actions of edit a exist properties
    if user_action == 'edit':
        if not has_perm('management.change_tcmsenvproperty'):
            return JsonResponse({'rc': 1, 'response': 'Permission denied'})

        if not request.GET.get('id'):
            return JsonResponse({'rc': 1, 'response': 'ID is required'})

        try:
            property_id = request.GET['id']
            env_property = TCMSEnvProperty.objects.get(id=int(property_id))
        except ValueError:
            return JsonResponse({
                'rc': 1,
                'response': f'ID {property_id} is not a valid integer.'
            })
        except TCMSEnvProperty.DoesNotExist:
            return JsonResponse({'rc': 1, 'response': 'ID does not exist.'})

        new_name = request.GET.get('name', env_property.name)
        env_property.name = new_name
        try:
            env_property.save(update_fields=['name'])
        except Exception:
            return JsonResponse({'rc': 1, 'response': 'Cannot save property.'})

        return JsonResponse({'rc': 0, 'response': 'ok'})

    # Actions of remove properties
    if user_action == 'modify':
        if not has_perm('management.change_tcmsenvproperty'):
            message = 'Permission denied'

        property_ids = request.GET.getlist('id')

        if has_perm('management.change_tcmsenvproperty') and property_ids:
            env_properties = TCMSEnvProperty.objects.filter(id__in=property_ids)

            if request.GET.get('status') in ['0', '1']:
                for env_property in env_properties:
                    env_property.is_active = int(request.GET['status'])
                    env_property.save()

                property_values = "', '".join(sorted(
                    env_properties.values_list('name', flat=True)))
                message = "Modify test properties status '%s' successfully." % property_values

                if not env_property.is_active:
                    TCMSEnvGroupPropertyMap.objects.filter(
                        property__id__in=property_ids).delete()
            else:
                message = 'Argument illegal'

    if request.is_ajax():
        ajax_response['rc'] = 1
        ajax_response['response'] = 'Unknown action'
        return HttpResponse(json_dumps(ajax_response))

    context_data = {
        'message': message,
        'properties': TCMSEnvProperty.objects.all().order_by('-is_active')
    }
    return render(request, template_name, context=context_data)


@require_GET
def environment_property_values(request):
    """
    List values of property
    """
    template_name = 'environment/ajax/property_values.html'
    message = ''
    duplicated_property_value = []

    if not request.GET.get('property_id'):
        return HttpResponse('Property ID should specify')

    try:
        property = TCMSEnvProperty.objects.get(id=request.GET['property_id'])
    except TCMSEnvProperty.DoesNotExist as error:
        return HttpResponse(error)

    user_action = request.GET.get('action')

    if user_action == 'add' and request.GET.get('value'):
        if not request.user.has_perm('management.add_tcmsenvvalue'):
            return HttpResponse('Permission denied')

        for value in request.GET['value'].split(','):
            try:
                property.value.create(value=value)
            except IntegrityError as error:
                if error[1].startswith('Duplicate'):
                    duplicated_property_value.append(value)

    if user_action == 'edit' and request.GET.get('id'):
        if not request.user.has_perm('management.change_tcmsenvvalue'):
            return HttpResponse('Permission denied')

        try:
            property_value = property.value.get(id=request.GET['id'])
            property_value.value = request.GET.get('value', property_value.value)
            try:
                property_value.save()
            except IntegrityError as error:
                if error[1].startswith('Duplicate'):
                    duplicated_property_value.append(property_value.value)

        except TCMSEnvValue.DoesNotExist as error:
            return HttpResponse(error[1])

    if user_action == 'modify' and request.GET.get('id'):
        if not request.user.has_perm('management.change_tcmsenvvalue'):
            return HttpResponse('Permission denied')

        values = property.value.filter(id__in=request.GET.getlist('id'))
        status = request.GET.get('status')
        if status in ['0', '1']:
            for value in values:
                value.is_active = int(status)
                value.save()
        else:
            return HttpResponse('Argument illegel')

    if duplicated_property_value:
        s = "', '".join(duplicated_property_value)
        message = (
            f"Value(s) named '{s}' already exists in this property, "
            f"please select another name."
        )

    values = property.value.all()
    context_data = {
        'property': property,
        'values': values,
        'message': message,
    }
    return render(request, template_name, context=context_data)



# -*- coding: utf-8 -*-
from django import template

from tcms.testcases.models import TestCasePlan

register = template.Library()


@register.filter(name='sortkey')
def sortkey(case, plan):
    try:
        tcp = TestCasePlan.objects.get(plan=plan, case=case)
    except TestCasePlan.DoesNotExist:
        return None
    else:
        return tcp.sortkey


@register.filter(name='testcaseplan')
def testcaseplan(case, plan):
    try:
        tcp = TestCasePlan.objects.get(plan=plan, case=case)
    except TestCasePlan.DoesNotExist:
        return None
    else:
        return tcp.pk
